import pytest

import pyteal as pt


def test_compile_single():
    expr = pt.Int(1)

    expected = """
#pragma version 2
int 1
return
""".strip()
    actual_application = pt.compileTeal(expr, pt.Mode.Application)
    actual_signature = pt.compileTeal(expr, pt.Mode.Signature)

    assert actual_application == actual_signature
    assert actual_application == expected


def test_compile_sequence():
    expr = pt.Seq([pt.Pop(pt.Int(1)), pt.Pop(pt.Int(2)), pt.Int(3) + pt.Int(4)])

    expected = """
#pragma version 2
int 1
pop
int 2
pop
int 3
int 4
+
return
""".strip()
    actual_application = pt.compileTeal(expr, pt.Mode.Application)
    actual_signature = pt.compileTeal(expr, pt.Mode.Signature)

    assert actual_application == actual_signature
    assert actual_application == expected


def test_compile_branch():
    expr = pt.If(pt.Int(1)).Then(pt.Int(2)).Else(pt.Int(3))

    expected = """
#pragma version 2
int 1
bnz main_l2
int 3
b main_l3
main_l2:
int 2
main_l3:
return
""".strip()
    actual_application = pt.compileTeal(expr, pt.Mode.Application)
    actual_signature = pt.compileTeal(expr, pt.Mode.Signature)

    assert actual_application == actual_signature
    assert actual_application == expected


def test_compile_branch_multiple():
    expr = (
        pt.If(pt.Int(1))
        .Then(pt.Int(2))
        .ElseIf(pt.Int(3))
        .Then(pt.Int(4))
        .Else(pt.Int(5))
    )

    expected = """
#pragma version 2
int 1
bnz main_l4
int 3
bnz main_l3
int 5
b main_l5
main_l3:
int 4
b main_l5
main_l4:
int 2
main_l5:
return
""".strip()
    actual_application = pt.compileTeal(expr, pt.Mode.Application)
    actual_signature = pt.compileTeal(expr, pt.Mode.Signature)

    assert actual_application == actual_signature
    assert actual_application == expected


def test_empty_branch():
    program = pt.Seq(
        [
            pt.If(pt.Txn.application_id() == pt.Int(0)).Then(pt.Seq()),
            pt.Approve(),
        ]
    )

    expected = """#pragma version 5
txn ApplicationID
int 0
==
bnz main_l1
main_l1:
int 1
return
    """.strip()
    actual = pt.compileTeal(
        program, pt.Mode.Application, version=5, assembleConstants=False
    )
    assert actual == expected


def test_compile_mode():
    expr = pt.App.globalGet(pt.Bytes("key"))

    expected = """
#pragma version 2
byte "key"
app_global_get
return
""".strip()
    actual_application = pt.compileTeal(expr, pt.Mode.Application)

    assert actual_application == expected

    with pytest.raises(pt.TealInputError):
        pt.compileTeal(expr, pt.Mode.Signature)


def test_compile_version_invalid():
    expr = pt.Int(1)

    with pytest.raises(pt.TealInputError):
        pt.compileTeal(expr, pt.Mode.Signature, version=1)  # too small

    with pytest.raises(pt.TealInputError):
        pt.compileTeal(expr, pt.Mode.Signature, version=8)  # too large

    with pytest.raises(pt.TealInputError):
        pt.compileTeal(expr, pt.Mode.Signature, version=2.0)  # decimal


def test_compile_version_2():
    expr = pt.Int(1)

    expected = """
#pragma version 2
int 1
return
""".strip()
    actual = pt.compileTeal(expr, pt.Mode.Signature, version=2)
    assert actual == expected


def test_compile_version_default():
    expr = pt.Int(1)

    actual_default = pt.compileTeal(expr, pt.Mode.Signature)
    actual_version_2 = pt.compileTeal(expr, pt.Mode.Signature, version=2)
    assert actual_default == actual_version_2


def test_compile_version_3():
    expr = pt.Int(1)

    expected = """
#pragma version 3
int 1
return
""".strip()
    actual = pt.compileTeal(expr, pt.Mode.Signature, version=3)
    assert actual == expected


def test_compile_version_4():
    expr = pt.Int(1)

    expected = """
#pragma version 4
int 1
return
""".strip()
    actual = pt.compileTeal(expr, pt.Mode.Signature, version=4)
    assert actual == expected


def test_compile_version_5():
    expr = pt.Int(1)
    expected = """
#pragma version 5
int 1
return
""".strip()
    actual = pt.compileTeal(expr, pt.Mode.Signature, version=5)
    assert actual == expected


def test_compile_version_6():
    expr = pt.Int(1)
    expected = """
#pragma version 6
int 1
return
""".strip()
    actual = pt.compileTeal(expr, pt.Mode.Signature, version=6)
    assert actual == expected


def test_slot_load_before_store():

    program = pt.AssetHolding.balance(pt.Int(0), pt.Int(0)).value()
    with pytest.raises(pt.TealInternalError):
        pt.compileTeal(program, pt.Mode.Application, version=2)

    program = pt.AssetHolding.balance(pt.Int(0), pt.Int(0)).hasValue()
    with pytest.raises(pt.TealInternalError):
        pt.compileTeal(program, pt.Mode.Application, version=2)

    program = pt.App.globalGetEx(pt.Int(0), pt.Bytes("key")).value()
    with pytest.raises(pt.TealInternalError):
        pt.compileTeal(program, pt.Mode.Application, version=2)

    program = pt.App.globalGetEx(pt.Int(0), pt.Bytes("key")).hasValue()
    with pytest.raises(pt.TealInternalError):
        pt.compileTeal(program, pt.Mode.Application, version=2)

    program = pt.ScratchVar().load()
    with pytest.raises(pt.TealInternalError):
        pt.compileTeal(program, pt.Mode.Application, version=2)


def test_assign_scratch_slots():
    myScratch = pt.ScratchVar(pt.TealType.uint64)
    otherScratch = pt.ScratchVar(pt.TealType.uint64, 1)
    anotherScratch = pt.ScratchVar(pt.TealType.uint64, 0)
    lastScratch = pt.ScratchVar(pt.TealType.uint64)
    prog = pt.Seq(
        [
            myScratch.store(pt.Int(5)),  # Slot 2
            otherScratch.store(pt.Int(0)),  # Slot 1
            anotherScratch.store(pt.Int(7)),  # Slot 0
            lastScratch.store(pt.Int(9)),  # Slot 3
            pt.Approve(),
        ]
    )

    expected = """
#pragma version 4
int 5
store 2
int 0
store 1
int 7
store 0
int 9
store 3
int 1
return
""".strip()
    actual = pt.compileTeal(prog, mode=pt.Mode.Signature, version=4)
    assert actual == expected


def test_scratchvar_double_assign_invalid():
    myvar = pt.ScratchVar(pt.TealType.uint64, 10)
    otherVar = pt.ScratchVar(pt.TealType.uint64, 10)
    prog = pt.Seq([myvar.store(pt.Int(5)), otherVar.store(pt.Int(0)), pt.Approve()])
    with pytest.raises(pt.TealInternalError):
        pt.compileTeal(prog, mode=pt.Mode.Signature, version=4)


def test_assembleConstants():
    program = pt.Itob(pt.Int(1) + pt.Int(1) + pt.Tmpl.Int("TMPL_VAR")) == pt.Concat(
        pt.Bytes("test"), pt.Bytes("test"), pt.Bytes("test2")
    )

    expectedNoAssemble = """
#pragma version 3
int 1
int 1
+
int TMPL_VAR
+
itob
byte "test"
byte "test"
concat
byte "test2"
concat
==
return
""".strip()
    actualNoAssemble = pt.compileTeal(
        program, pt.Mode.Application, version=3, assembleConstants=False
    )
    assert expectedNoAssemble == actualNoAssemble

    expectedAssemble = """
#pragma version 3
intcblock 1
bytecblock 0x74657374
intc_0 // 1
intc_0 // 1
+
pushint TMPL_VAR // TMPL_VAR
+
itob
bytec_0 // "test"
bytec_0 // "test"
concat
pushbytes 0x7465737432 // "test2"
concat
==
return
""".strip()
    actualAssemble = pt.compileTeal(
        program, pt.Mode.Application, version=3, assembleConstants=True
    )
    assert expectedAssemble == actualAssemble

    with pytest.raises(pt.TealInternalError):
        pt.compileTeal(program, pt.Mode.Application, version=2, assembleConstants=True)


def test_compile_while():
    i = pt.ScratchVar()
    program = pt.Seq(
        [
            i.store(pt.Int(0)),
            pt.While(i.load() < pt.Int(2)).Do(pt.Seq([i.store(i.load() + pt.Int(1))])),
            pt.Approve(),
        ]
    )

    expected = """
    #pragma version 4
int 0
store 0
main_l1:
load 0
int 2
<
bz main_l3
load 0
int 1
+
store 0
b main_l1
main_l3:
int 1
return
    """.strip()
    actual = pt.compileTeal(
        program, pt.Mode.Application, version=4, assembleConstants=False
    )
    assert expected == actual

    # nested
    i = pt.ScratchVar()
    j = pt.ScratchVar()

    program = pt.Seq(
        [
            i.store(pt.Int(0)),
            pt.While(i.load() < pt.Int(2)).Do(
                pt.Seq(
                    [
                        j.store(pt.Int(0)),
                        pt.While(j.load() < pt.Int(5)).Do(
                            pt.Seq([j.store(j.load() + pt.Int(1))])
                        ),
                        i.store(i.load() + pt.Int(1)),
                    ]
                )
            ),
            pt.Approve(),
        ]
    )

    expected = """#pragma version 4
int 0
store 0
main_l1:
load 0
int 2
<
bz main_l6
int 0
store 1
main_l3:
load 1
int 5
<
bnz main_l5
load 0
int 1
+
store 0
b main_l1
main_l5:
load 1
int 1
+
store 1
b main_l3
main_l6:
int 1
return
    """.strip()

    actual = pt.compileTeal(
        program, pt.Mode.Application, version=4, assembleConstants=False
    )
    assert expected == actual


def test_compile_for():
    i = pt.ScratchVar()
    program = pt.Seq(
        [
            pt.For(
                i.store(pt.Int(0)), i.load() < pt.Int(10), i.store(i.load() + pt.Int(1))
            ).Do(pt.Seq([pt.App.globalPut(pt.Itob(i.load()), i.load() * pt.Int(2))])),
            pt.Approve(),
        ]
    )

    expected = """#pragma version 4
int 0
store 0
main_l1:
load 0
int 10
<
bz main_l3
load 0
itob
load 0
int 2
*
app_global_put
load 0
int 1
+
store 0
b main_l1
main_l3:
int 1
return
    """.strip()
    actual = pt.compileTeal(
        program, pt.Mode.Application, version=4, assembleConstants=False
    )
    assert expected == actual

    # nested
    i = pt.ScratchVar()
    j = pt.ScratchVar()
    program = pt.Seq(
        [
            pt.For(
                i.store(pt.Int(0)), i.load() < pt.Int(10), i.store(i.load() + pt.Int(1))
            ).Do(
                pt.Seq(
                    [
                        pt.For(
                            j.store(pt.Int(0)),
                            j.load() < pt.Int(4),
                            j.store(j.load() + pt.Int(2)),
                        ).Do(
                            pt.Seq(
                                [
                                    pt.App.globalPut(
                                        pt.Itob(j.load()), j.load() * pt.Int(2)
                                    )
                                ]
                            )
                        )
                    ]
                )
            ),
            pt.Approve(),
        ]
    )

    expected = """#pragma version 4
int 0
store 0
main_l1:
load 0
int 10
<
bz main_l6
int 0
store 1
main_l3:
load 1
int 4
<
bnz main_l5
load 0
int 1
+
store 0
b main_l1
main_l5:
load 1
itob
load 1
int 2
*
app_global_put
load 1
int 2
+
store 1
b main_l3
main_l6:
int 1
return
    """.strip()
    actual = pt.compileTeal(
        program, pt.Mode.Application, version=4, assembleConstants=False
    )
    assert expected == actual


def test_compile_break():

    # pt.While
    i = pt.ScratchVar()
    program = pt.Seq(
        [
            i.store(pt.Int(0)),
            pt.While(i.load() < pt.Int(3)).Do(
                pt.Seq(
                    [
                        pt.If(i.load() == pt.Int(2), pt.Break()),
                        i.store(i.load() + pt.Int(1)),
                    ]
                )
            ),
            pt.Approve(),
        ]
    )

    expected = """#pragma version 4
int 0
store 0
main_l1:
load 0
int 3
<
bz main_l4
load 0
int 2
==
bnz main_l4
load 0
int 1
+
store 0
b main_l1
main_l4:
int 1
return
            """.strip()
    actual = pt.compileTeal(
        program, pt.Mode.Application, version=4, assembleConstants=False
    )
    assert expected == actual

    # pt.For
    i = pt.ScratchVar()
    program = pt.Seq(
        [
            pt.For(
                i.store(pt.Int(0)), i.load() < pt.Int(10), i.store(i.load() + pt.Int(1))
            ).Do(
                pt.Seq(
                    [
                        pt.If(i.load() == pt.Int(4), pt.Break()),
                        pt.App.globalPut(pt.Itob(i.load()), i.load() * pt.Int(2)),
                    ]
                )
            ),
            pt.Approve(),
        ]
    )

    expected = """#pragma version 4
int 0
store 0
main_l1:
load 0
int 10
<
bz main_l4
load 0
int 4
==
bnz main_l4
load 0
itob
load 0
int 2
*
app_global_put
load 0
int 1
+
store 0
b main_l1
main_l4:
int 1
return
        """.strip()
    actual = pt.compileTeal(
        program, pt.Mode.Application, version=4, assembleConstants=False
    )
    assert expected == actual


def test_compile_continue():
    # pt.While
    i = pt.ScratchVar()
    program = pt.Seq(
        [
            i.store(pt.Int(0)),
            pt.While(i.load() < pt.Int(3)).Do(
                pt.Seq(
                    [
                        pt.If(i.load() == pt.Int(2), pt.Continue()),
                        i.store(i.load() + pt.Int(1)),
                    ]
                )
            ),
            pt.Approve(),
        ]
    )

    expected = """#pragma version 4
int 0
store 0
main_l1:
load 0
int 3
<
bz main_l4
load 0
int 2
==
bnz main_l1
load 0
int 1
+
store 0
b main_l1
main_l4:
int 1
return
    """.strip()
    actual = pt.compileTeal(
        program, pt.Mode.Application, version=4, assembleConstants=False
    )
    assert expected == actual

    # pt.While
    program = pt.Seq(
        i.store(pt.Int(0)),
        pt.While(i.load() < pt.Int(30)).Do(
            pt.Seq(
                i.store(i.load() + pt.Int(1)),
                pt.Continue(),
            )
        ),
        pt.Return(pt.Int(1)),
    )

    expected = """#pragma version 4
int 0
store 0
main_l1:
load 0
int 30
<
bz main_l3
load 0
int 1
+
store 0
b main_l1
main_l3:
int 1
return
    """.strip()
    actual = pt.compileTeal(
        program, pt.Mode.Application, version=4, assembleConstants=False
    )
    assert expected == actual

    # pt.For
    i = pt.ScratchVar()
    program = pt.Seq(
        [
            pt.For(
                i.store(pt.Int(0)), i.load() < pt.Int(10), i.store(i.load() + pt.Int(1))
            ).Do(
                pt.Seq(
                    [
                        pt.If(i.load() == pt.Int(4), pt.Continue()),
                        pt.App.globalPut(pt.Itob(i.load()), i.load() * pt.Int(2)),
                    ]
                )
            ),
            pt.Approve(),
        ]
    )

    expected = """#pragma version 4
int 0
store 0
main_l1:
load 0
int 10
<
bz main_l5
load 0
int 4
==
bnz main_l4
load 0
itob
load 0
int 2
*
app_global_put
main_l4:
load 0
int 1
+
store 0
b main_l1
main_l5:
int 1
return
            """.strip()
    actual = pt.compileTeal(
        program, pt.Mode.Application, version=4, assembleConstants=False
    )
    assert expected == actual


def test_compile_continue_break_nested():

    i = pt.ScratchVar()
    program = pt.Seq(
        [
            i.store(pt.Int(0)),
            pt.While(i.load() < pt.Int(10)).Do(
                pt.Seq(
                    [
                        i.store(i.load() + pt.Int(1)),
                        pt.If(i.load() < pt.Int(4), pt.Continue(), pt.Break()),
                    ]
                )
            ),
            pt.Approve(),
        ]
    )

    expected = """#pragma version 4
int 0
store 0
main_l1:
load 0
int 10
<
bz main_l3
load 0
int 1
+
store 0
load 0
int 4
<
bnz main_l1
main_l3:
int 1
return
    """.strip()
    actual = pt.compileTeal(
        program, pt.Mode.Application, version=4, assembleConstants=False
    )
    assert expected == actual

    i = pt.ScratchVar()
    program = pt.Seq(
        [
            i.store(pt.Int(0)),
            pt.While(i.load() < pt.Int(10)).Do(
                pt.Seq(
                    [
                        pt.If(i.load() == pt.Int(8), pt.Break()),
                        pt.While(i.load() < pt.Int(6)).Do(
                            pt.Seq(
                                [
                                    pt.If(i.load() == pt.Int(3), pt.Break()),
                                    i.store(i.load() + pt.Int(1)),
                                ]
                            )
                        ),
                        pt.If(i.load() < pt.Int(5), pt.Continue()),
                        i.store(i.load() + pt.Int(1)),
                    ]
                )
            ),
            pt.Approve(),
        ]
    )

    expected = """#pragma version 4
int 0
store 0
main_l1:
load 0
int 10
<
bz main_l8
load 0
int 8
==
bnz main_l8
main_l3:
load 0
int 6
<
bnz main_l6
main_l4:
load 0
int 5
<
bnz main_l1
load 0
int 1
+
store 0
b main_l1
main_l6:
load 0
int 3
==
bnz main_l4
load 0
int 1
+
store 0
b main_l3
main_l8:
int 1
return
""".strip()
    actual = pt.compileTeal(
        program, pt.Mode.Application, version=4, assembleConstants=False
    )
    assert expected == actual


def test_compile_subroutine_unsupported():
    @pt.Subroutine(pt.TealType.none)
    def storeValue(value: pt.Expr) -> pt.Expr:
        return pt.App.globalPut(pt.Bytes("key"), value)

    program = pt.Seq(
        [
            pt.If(pt.Txn.sender() == pt.Global.creator_address()).Then(
                storeValue(pt.Txn.application_args[0])
            ),
            pt.Approve(),
        ]
    )

    with pytest.raises(pt.TealInputError):
        pt.compileTeal(program, pt.Mode.Application, version=3)


def test_compile_subroutine_no_return():
    @pt.Subroutine(pt.TealType.none)
    def storeValue(value: pt.Expr) -> pt.Expr:
        return pt.App.globalPut(pt.Bytes("key"), value)

    program = pt.Seq(
        [
            pt.If(pt.Txn.sender() == pt.Global.creator_address()).Then(
                storeValue(pt.Txn.application_args[0])
            ),
            pt.Approve(),
        ]
    )

    expected = """#pragma version 4
txn Sender
global CreatorAddress
==
bz main_l2
txna ApplicationArgs 0
callsub storeValue_0
main_l2:
int 1
return

// storeValue
storeValue_0:
store 0
byte "key"
load 0
app_global_put
retsub
    """.strip()
    actual = pt.compileTeal(
        program, pt.Mode.Application, version=4, assembleConstants=False
    )
    assert actual == expected


def test_compile_subroutine_with_return():
    @pt.Subroutine(pt.TealType.none)
    def storeValue(value: pt.Expr) -> pt.Expr:
        return pt.App.globalPut(pt.Bytes("key"), value)

    @pt.Subroutine(pt.TealType.bytes)
    def getValue() -> pt.Expr:
        return pt.App.globalGet(pt.Bytes("key"))

    program = pt.Seq(
        [
            pt.If(pt.Txn.sender() == pt.Global.creator_address()).Then(
                storeValue(pt.Txn.application_args[0])
            ),
            pt.If(getValue() == pt.Bytes("fail")).Then(pt.Reject()),
            pt.Approve(),
        ]
    )

    expected = """#pragma version 4
txn Sender
global CreatorAddress
==
bnz main_l3
main_l1:
callsub getValue_1
byte "fail"
==
bz main_l4
int 0
return
main_l3:
txna ApplicationArgs 0
callsub storeValue_0
b main_l1
main_l4:
int 1
return

// storeValue
storeValue_0:
store 0
byte "key"
load 0
app_global_put
retsub

// getValue
getValue_1:
byte "key"
app_global_get
retsub
    """.strip()
    actual = pt.compileTeal(
        program, pt.Mode.Application, version=4, assembleConstants=False
    )
    assert actual == expected


def test_compile_subroutine_many_args():
    @pt.Subroutine(pt.TealType.uint64)
    def calculateSum(
        a1: pt.Expr, a2: pt.Expr, a3: pt.Expr, a4: pt.Expr, a5: pt.Expr, a6: pt.Expr
    ) -> pt.Expr:
        return a1 + a2 + a3 + a4 + a5 + a6

    program = pt.Return(
        calculateSum(pt.Int(1), pt.Int(2), pt.Int(3), pt.Int(4), pt.Int(5), pt.Int(6))
        == pt.Int(1 + 2 + 3 + 4 + 5 + 6)
    )

    expected = """#pragma version 4
int 1
int 2
int 3
int 4
int 5
int 6
callsub calculateSum_0
int 21
==
return

// calculateSum
calculateSum_0:
store 5
store 4
store 3
store 2
store 1
store 0
load 0
load 1
+
load 2
+
load 3
+
load 4
+
load 5
+
retsub
    """.strip()
    actual = pt.compileTeal(
        program, pt.Mode.Application, version=4, assembleConstants=False
    )
    assert actual == expected


def test_compile_subroutine_recursive():
    @pt.Subroutine(pt.TealType.uint64)
    def isEven(i: pt.Expr) -> pt.Expr:
        return (
            pt.If(i == pt.Int(0))
            .Then(pt.Int(1))
            .ElseIf(i == pt.Int(1))
            .Then(pt.Int(0))
            .Else(isEven(i - pt.Int(2)))
        )

    program = pt.Return(isEven(pt.Int(6)))

    expected = """#pragma version 4
int 6
callsub isEven_0
return

// isEven
isEven_0:
store 0
load 0
int 0
==
bnz isEven_0_l4
load 0
int 1
==
bnz isEven_0_l3
load 0
int 2
-
load 0
dig 1
callsub isEven_0
swap
store 0
swap
pop
b isEven_0_l5
isEven_0_l3:
int 0
b isEven_0_l5
isEven_0_l4:
int 1
isEven_0_l5:
retsub
    """.strip()
    actual = pt.compileTeal(
        program, pt.Mode.Application, version=4, assembleConstants=False
    )
    assert actual == expected


def test_compile_subroutine_recursive_5():
    @pt.Subroutine(pt.TealType.uint64)
    def isEven(i: pt.Expr) -> pt.Expr:
        return (
            pt.If(i == pt.Int(0))
            .Then(pt.Int(1))
            .ElseIf(i == pt.Int(1))
            .Then(pt.Int(0))
            .Else(isEven(i - pt.Int(2)))
        )

    program = pt.Return(isEven(pt.Int(6)))

    expected = """#pragma version 5
int 6
callsub isEven_0
return

// isEven
isEven_0:
store 0
load 0
int 0
==
bnz isEven_0_l4
load 0
int 1
==
bnz isEven_0_l3
load 0
int 2
-
load 0
swap
callsub isEven_0
swap
store 0
b isEven_0_l5
isEven_0_l3:
int 0
b isEven_0_l5
isEven_0_l4:
int 1
isEven_0_l5:
retsub
    """.strip()
    actual = pt.compileTeal(
        program, pt.Mode.Application, version=5, assembleConstants=False
    )
    assert actual == expected


def test_compile_subroutine_recursive_multiple_args():
    @pt.Subroutine(pt.TealType.uint64)
    def multiplyByAdding(a, b):
        return (
            pt.If(a == pt.Int(0))
            .Then(pt.Return(pt.Int(0)))
            .Else(pt.Return(b + multiplyByAdding(a - pt.Int(1), b)))
        )

    program = pt.Return(multiplyByAdding(pt.Int(3), pt.Int(5)))

    expected = """#pragma version 4
int 3
int 5
callsub multiplyByAdding_0
return

// multiplyByAdding
multiplyByAdding_0:
store 1
store 0
load 0
int 0
==
bnz multiplyByAdding_0_l2
load 1
load 0
int 1
-
load 1
load 0
load 1
dig 3
dig 3
callsub multiplyByAdding_0
store 0
store 1
load 0
swap
store 0
swap
pop
swap
pop
+
retsub
multiplyByAdding_0_l2:
int 0
retsub
    """.strip()
    actual = pt.compileTeal(
        program, pt.Mode.Application, version=4, assembleConstants=False
    )
    assert actual == expected


def test_compile_subroutine_recursive_multiple_args_5():
    @pt.Subroutine(pt.TealType.uint64)
    def multiplyByAdding(a, b):
        return (
            pt.If(a == pt.Int(0))
            .Then(pt.Return(pt.Int(0)))
            .Else(pt.Return(b + multiplyByAdding(a - pt.Int(1), b)))
        )

    program = pt.Return(multiplyByAdding(pt.Int(3), pt.Int(5)))

    expected = """#pragma version 5
int 3
int 5
callsub multiplyByAdding_0
return

// multiplyByAdding
multiplyByAdding_0:
store 1
store 0
load 0
int 0
==
bnz multiplyByAdding_0_l2
load 1
load 0
int 1
-
load 1
load 0
load 1
uncover 3
uncover 3
callsub multiplyByAdding_0
cover 2
store 1
store 0
+
retsub
multiplyByAdding_0_l2:
int 0
retsub
    """.strip()
    actual = pt.compileTeal(
        program, pt.Mode.Application, version=5, assembleConstants=False
    )
    assert actual == expected


def test_compile_subroutine_mutually_recursive_4():
    @pt.Subroutine(pt.TealType.uint64)
    def isEven(i: pt.Expr) -> pt.Expr:
        return pt.If(i == pt.Int(0), pt.Int(1), pt.Not(isOdd(i - pt.Int(1))))

    @pt.Subroutine(pt.TealType.uint64)
    def isOdd(i: pt.Expr) -> pt.Expr:
        return pt.If(i == pt.Int(0), pt.Int(0), pt.Not(isEven(i - pt.Int(1))))

    program = pt.Return(isEven(pt.Int(6)))

    expected = """#pragma version 4
int 6
callsub isEven_0
return

// isEven
isEven_0:
store 0
load 0
int 0
==
bnz isEven_0_l2
load 0
int 1
-
load 0
dig 1
callsub isOdd_1
swap
store 0
swap
pop
!
b isEven_0_l3
isEven_0_l2:
int 1
isEven_0_l3:
retsub

// isOdd
isOdd_1:
store 1
load 1
int 0
==
bnz isOdd_1_l2
load 1
int 1
-
load 1
dig 1
callsub isEven_0
swap
store 1
swap
pop
!
b isOdd_1_l3
isOdd_1_l2:
int 0
isOdd_1_l3:
retsub
    """.strip()
    actual = pt.compileTeal(
        program, pt.Mode.Application, version=4, assembleConstants=False
    )
    assert actual == expected


def test_compile_subroutine_mutually_recursive_5():
    @pt.Subroutine(pt.TealType.uint64)
    def isEven(i: pt.Expr) -> pt.Expr:
        return pt.If(i == pt.Int(0), pt.Int(1), pt.Not(isOdd(i - pt.Int(1))))

    @pt.Subroutine(pt.TealType.uint64)
    def isOdd(i: pt.Expr) -> pt.Expr:
        return pt.If(i == pt.Int(0), pt.Int(0), pt.Not(isEven(i - pt.Int(1))))

    program = pt.Return(isEven(pt.Int(6)))

    expected = """#pragma version 5
int 6
callsub isEven_0
return

// isEven
isEven_0:
store 0
load 0
int 0
==
bnz isEven_0_l2
load 0
int 1
-
load 0
swap
callsub isOdd_1
swap
store 0
!
b isEven_0_l3
isEven_0_l2:
int 1
isEven_0_l3:
retsub

// isOdd
isOdd_1:
store 1
load 1
int 0
==
bnz isOdd_1_l2
load 1
int 1
-
load 1
swap
callsub isEven_0
swap
store 1
!
b isOdd_1_l3
isOdd_1_l2:
int 0
isOdd_1_l3:
retsub
    """.strip()
    actual = pt.compileTeal(
        program, pt.Mode.Application, version=5, assembleConstants=False
    )
    assert actual == expected


def test_compile_subroutine_mutually_recursive_different_arg_count_4():
    @pt.Subroutine(pt.TealType.uint64)
    def factorial(i: pt.Expr) -> pt.Expr:
        return pt.If(
            i <= pt.Int(1),
            pt.Int(1),
            factorial_intermediate(i - pt.Int(1), pt.Bytes("inconsequential")) * i,
        )

    @pt.Subroutine(pt.TealType.uint64)
    def factorial_intermediate(i: pt.Expr, j: pt.Expr) -> pt.Expr:
        return pt.Seq(pt.Pop(j), factorial(i))

    program = pt.Return(factorial(pt.Int(4)) == pt.Int(24))

    expected = """#pragma version 4
int 4
callsub factorial_0
int 24
==
return

// factorial
factorial_0:
store 0
load 0
int 1
<=
bnz factorial_0_l2
load 0
int 1
-
byte "inconsequential"
load 0
dig 2
dig 2
callsub factorialintermediate_1
swap
store 0
swap
pop
swap
pop
load 0
*
b factorial_0_l3
factorial_0_l2:
int 1
factorial_0_l3:
retsub

// factorial_intermediate
factorialintermediate_1:
store 2
store 1
load 2
pop
load 1
load 1
load 2
dig 2
callsub factorial_0
store 1
store 2
load 1
swap
store 1
swap
pop
retsub
    """.strip()
    actual = pt.compileTeal(
        program, pt.Mode.Application, version=4, assembleConstants=False
    )
    assert actual == expected


def test_compile_subroutine_mutually_recursive_different_arg_count_5():
    @pt.Subroutine(pt.TealType.uint64)
    def factorial(i: pt.Expr) -> pt.Expr:
        return pt.If(
            i <= pt.Int(1),
            pt.Int(1),
            factorial_intermediate(i - pt.Int(1), pt.Bytes("inconsequential")) * i,
        )

    @pt.Subroutine(pt.TealType.uint64)
    def factorial_intermediate(i: pt.Expr, j: pt.Expr) -> pt.Expr:
        return pt.Seq(pt.Log(j), factorial(i))

    program = pt.Return(factorial(pt.Int(4)) == pt.Int(24))

    expected = """#pragma version 5
int 4
callsub factorial_0
int 24
==
return

// factorial
factorial_0:
store 0
load 0
int 1
<=
bnz factorial_0_l2
load 0
int 1
-
byte "inconsequential"
load 0
cover 2
callsub factorialintermediate_1
swap
store 0
load 0
*
b factorial_0_l3
factorial_0_l2:
int 1
factorial_0_l3:
retsub

// factorial_intermediate
factorialintermediate_1:
store 2
store 1
load 2
log
load 1
load 1
load 2
uncover 2
callsub factorial_0
cover 2
store 2
store 1
retsub
    """.strip()
    actual = pt.compileTeal(
        program, pt.Mode.Application, version=5, assembleConstants=False
    )
    assert actual == expected


def test_compile_loop_in_subroutine():
    @pt.Subroutine(pt.TealType.none)
    def setState(value: pt.Expr) -> pt.Expr:
        i = pt.ScratchVar()
        return pt.For(
            i.store(pt.Int(0)), i.load() < pt.Int(10), i.store(i.load() + pt.Int(1))
        ).Do(pt.App.globalPut(pt.Itob(i.load()), value))

    program = pt.Seq([setState(pt.Bytes("value")), pt.Approve()])

    expected = """#pragma version 4
byte "value"
callsub setState_0
int 1
return

// setState
setState_0:
store 0
int 0
store 1
setState_0_l1:
load 1
int 10
<
bz setState_0_l3
load 1
itob
load 0
app_global_put
load 1
int 1
+
store 1
b setState_0_l1
setState_0_l3:
retsub
    """.strip()
    actual = pt.compileTeal(
        program, pt.Mode.Application, version=4, assembleConstants=False
    )
    assert actual == expected


def test_compile_subroutine_invalid_name():
    def tmp() -> pt.Expr:
        return pt.Int(1)

    tmp.__name__ = "invalid-;)"

    program = pt.Subroutine(pt.TealType.uint64)(tmp)()
    expected = """#pragma version 4
callsub invalid_0
return

// invalid-;)
invalid_0:
int 1
retsub
    """.strip()
    actual = pt.compileTeal(
        program, pt.Mode.Application, version=4, assembleConstants=False
    )
    assert actual == expected


def test_compile_subroutine_assemble_constants():
    @pt.Subroutine(pt.TealType.none)
    def storeValue(key: pt.Expr, t1: pt.Expr, t2: pt.Expr, t3: pt.Expr) -> pt.Expr:
        return pt.App.globalPut(key, t1 + t2 + t3 + pt.Int(10))

    program = pt.Seq(
        [
            pt.If(pt.Txn.application_id() == pt.Int(0)).Then(
                storeValue(
                    pt.Concat(pt.Bytes("test"), pt.Bytes("test"), pt.Bytes("a")),
                    pt.Int(1),
                    pt.Int(1),
                    pt.Int(3),
                )
            ),
            pt.Approve(),
        ]
    )

    expected = """#pragma version 4
intcblock 1
bytecblock 0x74657374
txn ApplicationID
pushint 0 // 0
==
bz main_l2
bytec_0 // "test"
bytec_0 // "test"
concat
pushbytes 0x61 // "a"
concat
intc_0 // 1
intc_0 // 1
pushint 3 // 3
callsub storeValue_0
main_l2:
intc_0 // 1
return

// storeValue
storeValue_0:
store 3
store 2
store 1
store 0
load 0
load 1
load 2
+
load 3
+
pushint 10 // 10
+
app_global_put
retsub
    """.strip()
    actual = pt.compileTeal(
        program, pt.Mode.Application, version=4, assembleConstants=True
    )
    assert actual == expected


def test_compile_subroutine_deferred_expr():
    @pt.Subroutine(pt.TealType.none)
    def deferredExample(value: pt.Expr) -> pt.Expr:
        return pt.Seq(
            pt.If(value == pt.Int(0)).Then(pt.Return()),
            pt.If(value == pt.Int(1)).Then(pt.Approve()),
            pt.If(value == pt.Int(2)).Then(pt.Reject()),
            pt.If(value == pt.Int(3)).Then(pt.Err()),
        )

    program = pt.Seq(deferredExample(pt.Int(10)), pt.Approve())

    expected_no_deferred = """#pragma version 6
int 10
callsub deferredExample_0
int 1
return

// deferredExample
deferredExample_0:
store 0
load 0
int 0
==
bnz deferredExample_0_l7
load 0
int 1
==
bnz deferredExample_0_l6
load 0
int 2
==
bnz deferredExample_0_l5
load 0
int 3
==
bz deferredExample_0_l8
err
deferredExample_0_l5:
int 0
return
deferredExample_0_l6:
int 1
return
deferredExample_0_l7:
retsub
deferredExample_0_l8:
retsub
    """.strip()
    actual_no_deferred = pt.compileTeal(
        program, pt.Mode.Application, version=6, assembleConstants=False
    )
    assert actual_no_deferred == expected_no_deferred

    # manually add deferred expression to SubroutineDefinition
    declaration = deferredExample.subroutine.get_declaration()
    declaration.deferred_expr = pt.Pop(pt.Bytes("deferred"))

    expected_deferred = """#pragma version 6
int 10
callsub deferredExample_0
int 1
return

// deferredExample
deferredExample_0:
store 0
load 0
int 0
==
bnz deferredExample_0_l7
load 0
int 1
==
bnz deferredExample_0_l6
load 0
int 2
==
bnz deferredExample_0_l5
load 0
int 3
==
bz deferredExample_0_l8
err
deferredExample_0_l5:
int 0
return
deferredExample_0_l6:
int 1
return
deferredExample_0_l7:
byte "deferred"
pop
retsub
deferredExample_0_l8:
byte "deferred"
pop
retsub
    """.strip()
    actual_deferred = pt.compileTeal(
        program, pt.Mode.Application, version=6, assembleConstants=False
    )
    assert actual_deferred == expected_deferred


def test_compile_subroutine_deferred_expr_empty():
    @pt.Subroutine(pt.TealType.none)
    def empty() -> pt.Expr:
        return pt.Return()

    program = pt.Seq(empty(), pt.Approve())

    expected_no_deferred = """#pragma version 6
callsub empty_0
int 1
return

// empty
empty_0:
retsub
    """.strip()
    actual_no_deferred = pt.compileTeal(
        program, pt.Mode.Application, version=6, assembleConstants=False
    )
    assert actual_no_deferred == expected_no_deferred

    # manually add deferred expression to SubroutineDefinition
    declaration = empty.subroutine.get_declaration()
    declaration.deferred_expr = pt.Pop(pt.Bytes("deferred"))

    expected_deferred = """#pragma version 6
callsub empty_0
int 1
return

// empty
empty_0:
byte "deferred"
pop
retsub
    """.strip()
    actual_deferred = pt.compileTeal(
        program, pt.Mode.Application, version=6, assembleConstants=False
    )
    assert actual_deferred == expected_deferred


def test_compileSubroutine_deferred_block_malformed():
    class BadRetsub(pt.Expr):
        def type_of(self) -> pt.TealType:
            return pt.TealType.none

        def has_return(self) -> bool:
            return True

        def __str__(self) -> str:
            return "(BadRetsub)"

        def __teal__(
            self, options: pt.CompileOptions
        ) -> tuple[pt.TealBlock, pt.TealSimpleBlock]:
            block = pt.TealSimpleBlock(
                [
                    pt.TealOp(self, pt.Op.int, 1),
                    pt.TealOp(self, pt.Op.pop),
                    pt.TealOp(self, pt.Op.retsub),
                ]
            )

            return block, block

    @pt.Subroutine(pt.TealType.none)
    def bad() -> pt.Expr:
        return BadRetsub()

    program = pt.Seq(bad(), pt.Approve())

    # manually add deferred expression to SubroutineDefinition
    declaration = bad.subroutine.get_declaration()
    declaration.deferred_expr = pt.Pop(pt.Bytes("deferred"))

    with pytest.raises(
        pt.TealInternalError,
        match=r"^Expected retsub to be the only op in the block, but there are 3 ops$",
    ):
        pt.compileTeal(program, pt.Mode.Application, version=6, assembleConstants=False)


def test_compile_wide_ratio():
    cases = (
        (
            pt.WideRatio([pt.Int(2), pt.Int(100)], [pt.Int(5)]),
            """#pragma version 5
int 2
int 100
mulw
int 0
int 5
divmodw
pop
pop
swap
!
assert
return
""",
        ),
        (
            pt.WideRatio([pt.Int(2), pt.Int(100)], [pt.Int(10), pt.Int(5)]),
            """#pragma version 5
int 2
int 100
mulw
int 10
int 5
mulw
divmodw
pop
pop
swap
!
assert
return
""",
        ),
        (
            pt.WideRatio([pt.Int(2), pt.Int(100), pt.Int(3)], [pt.Int(10), pt.Int(5)]),
            """#pragma version 5
int 2
int 100
mulw
int 3
uncover 2
dig 1
*
cover 2
mulw
cover 2
+
swap
int 10
int 5
mulw
divmodw
pop
pop
swap
!
assert
return
""",
        ),
        (
            pt.WideRatio(
                [pt.Int(2), pt.Int(100), pt.Int(3)], [pt.Int(10), pt.Int(5), pt.Int(6)]
            ),
            """#pragma version 5
int 2
int 100
mulw
int 3
uncover 2
dig 1
*
cover 2
mulw
cover 2
+
swap
int 10
int 5
mulw
int 6
uncover 2
dig 1
*
cover 2
mulw
cover 2
+
swap
divmodw
pop
pop
swap
!
assert
return
""",
        ),
        (
            pt.WideRatio(
                [pt.Int(2), pt.Int(100), pt.Int(3), pt.Int(4)],
                [pt.Int(10), pt.Int(5), pt.Int(6)],
            ),
            """#pragma version 5
int 2
int 100
mulw
int 3
uncover 2
dig 1
*
cover 2
mulw
cover 2
+
swap
int 4
uncover 2
dig 1
*
cover 2
mulw
cover 2
+
swap
int 10
int 5
mulw
int 6
uncover 2
dig 1
*
cover 2
mulw
cover 2
+
swap
divmodw
pop
pop
swap
!
assert
return
""",
        ),
    )

    for program, expected in cases:
        actual = pt.compileTeal(
            program, pt.Mode.Application, version=5, assembleConstants=False
        )
        assert actual == expected.strip()


def test_compile_abi_subroutine_return():
    @pt.ABIReturnSubroutine
    def abi_sum(
        toSum: pt.abi.DynamicArray[pt.abi.Uint64], *, output: pt.abi.Uint64
    ) -> pt.Expr:
        i = pt.ScratchVar(pt.TealType.uint64)
        valueAtIndex = pt.abi.Uint64()
        return pt.Seq(
            output.set(0),
            pt.For(
                i.store(pt.Int(0)),
                i.load() < toSum.length(),
                i.store(i.load() + pt.Int(1)),
            ).Do(
                pt.Seq(
                    toSum[i.load()].store_into(valueAtIndex),
                    output.set(output.get() + valueAtIndex.get()),
                )
            ),
        )

    program = pt.Seq(
        (to_sum_arr := pt.abi.make(pt.abi.DynamicArray[pt.abi.Uint64])).decode(
            pt.Txn.application_args[1]
        ),
        (res := pt.abi.Uint64()).set(abi_sum(to_sum_arr)),
        pt.abi.MethodReturn(res),
        pt.Approve(),
    )

    expected_sum = """#pragma version 6
txna ApplicationArgs 1
store 0
load 0
callsub abisum_0
store 1
byte 0x151f7c75
load 1
itob
concat
log
int 1
return

// abi_sum
abisum_0:
store 2
int 0
store 3
int 0
store 4
abisum_0_l1:
load 4
load 2
int 0
extract_uint16
store 6
load 6
<
bz abisum_0_l3
load 2
int 8
load 4
*
int 2
+
extract_uint64
store 5
load 3
load 5
+
store 3
load 4
int 1
+
store 4
b abisum_0_l1
abisum_0_l3:
load 3
retsub
    """.strip()

    actual_sum = pt.compileTeal(program, pt.Mode.Application, version=6)
    assert expected_sum == actual_sum

    @pt.ABIReturnSubroutine
    def conditional_factorial(
        _factor: pt.abi.Uint64, *, output: pt.abi.Uint64
    ) -> pt.Expr:
        i = pt.ScratchVar(pt.TealType.uint64)

        return pt.Seq(
            output.set(1),
            pt.If(_factor.get() <= pt.Int(1))
            .Then(pt.Return())
            .Else(
                pt.For(
                    i.store(_factor.get()),
                    i.load() > pt.Int(1),
                    i.store(i.load() - pt.Int(1)),
                ).Do(output.set(output.get() * i.load())),
            ),
        )

    program_cond_factorial = pt.Seq(
        (factor := pt.abi.Uint64()).decode(pt.Txn.application_args[1]),
        (res := pt.abi.Uint64()).set(conditional_factorial(factor)),
        pt.abi.MethodReturn(res),
        pt.Approve(),
    )

    expected_conditional_factorial = """#pragma version 6
txna ApplicationArgs 1
btoi
store 0
load 0
callsub conditionalfactorial_0
store 1
byte 0x151f7c75
load 1
itob
concat
log
int 1
return

// conditional_factorial
conditionalfactorial_0:
store 2
int 1
store 3
load 2
int 1
<=
bnz conditionalfactorial_0_l4
load 2
store 4
conditionalfactorial_0_l2:
load 4
int 1
>
bz conditionalfactorial_0_l5
load 3
load 4
*
store 3
load 4
int 1
-
store 4
b conditionalfactorial_0_l2
conditionalfactorial_0_l4:
load 3
retsub
conditionalfactorial_0_l5:
load 3
retsub
    """.strip()

    actual_conditional_factorial = pt.compileTeal(
        program_cond_factorial, pt.Mode.Application, version=6
    )
    assert actual_conditional_factorial == expected_conditional_factorial

    @pt.ABIReturnSubroutine
    def load_b4_set(*, output: pt.abi.Bool):
        return pt.Return()

    program_load_b4_set_broken = pt.Seq(
        (_ := pt.abi.Bool()).set(load_b4_set()), pt.Approve()
    )

    with pytest.raises(pt.TealInternalError):
        pt.compileTeal(program_load_b4_set_broken, pt.Mode.Application, version=6)

    @pt.ABIReturnSubroutine
    def access_b4_store(magic_num: pt.abi.Uint64, *, output: pt.abi.Uint64):
        return pt.Seq(output.set(output.get() ^ magic_num.get()))

    program_access_b4_store_broken = pt.Seq(
        (other_party_magic := pt.abi.Uint64()).decode(pt.Txn.application_args[1]),
        (_ := pt.abi.Uint64()).set(access_b4_store(other_party_magic)),
        pt.Approve(),
    )

    with pytest.raises(pt.TealInternalError):
        pt.compileTeal(program_access_b4_store_broken, pt.Mode.Application, version=6)


def test_router_app():
    def add_methods_to_router(router: pt.Router):
        @pt.ABIReturnSubroutine
        def add(
            a: pt.abi.Uint64, b: pt.abi.Uint64, *, output: pt.abi.Uint64
        ) -> pt.Expr:
            return output.set(a.get() + b.get())

        meth = router.add_method_handler(add)
        assert meth.method_signature() == "add(uint64,uint64)uint64"

        @pt.ABIReturnSubroutine
        def sub(
            a: pt.abi.Uint64, b: pt.abi.Uint64, *, output: pt.abi.Uint64
        ) -> pt.Expr:
            return output.set(a.get() - b.get())

        meth = router.add_method_handler(sub)
        assert meth.method_signature() == "sub(uint64,uint64)uint64"

        @pt.ABIReturnSubroutine
        def mul(
            a: pt.abi.Uint64, b: pt.abi.Uint64, *, output: pt.abi.Uint64
        ) -> pt.Expr:
            return output.set(a.get() * b.get())

        meth = router.add_method_handler(mul)
        assert meth.method_signature() == "mul(uint64,uint64)uint64"

        @pt.ABIReturnSubroutine
        def div(
            a: pt.abi.Uint64, b: pt.abi.Uint64, *, output: pt.abi.Uint64
        ) -> pt.Expr:
            return output.set(a.get() / b.get())

        meth = router.add_method_handler(div)
        assert meth.method_signature() == "div(uint64,uint64)uint64"

        @pt.ABIReturnSubroutine
        def mod(
            a: pt.abi.Uint64, b: pt.abi.Uint64, *, output: pt.abi.Uint64
        ) -> pt.Expr:
            return output.set(a.get() % b.get())

        meth = router.add_method_handler(mod)
        assert meth.method_signature() == "mod(uint64,uint64)uint64"

        @pt.ABIReturnSubroutine
        def all_laid_to_args(
            _a: pt.abi.Uint64,
            _b: pt.abi.Uint64,
            _c: pt.abi.Uint64,
            _d: pt.abi.Uint64,
            _e: pt.abi.Uint64,
            _f: pt.abi.Uint64,
            _g: pt.abi.Uint64,
            _h: pt.abi.Uint64,
            _i: pt.abi.Uint64,
            _j: pt.abi.Uint64,
            _k: pt.abi.Uint64,
            _l: pt.abi.Uint64,
            _m: pt.abi.Uint64,
            _n: pt.abi.Uint64,
            _o: pt.abi.Uint64,
            _p: pt.abi.Uint64,
            *,
            output: pt.abi.Uint64,
        ):
            return output.set(
                _a.get()
                + _b.get()
                + _c.get()
                + _d.get()
                + _e.get()
                + _f.get()
                + _g.get()
                + _h.get()
                + _i.get()
                + _j.get()
                + _k.get()
                + _l.get()
                + _m.get()
                + _n.get()
                + _o.get()
                + _p.get()
            )

        meth = router.add_method_handler(all_laid_to_args)
        assert (
            meth.method_signature()
            == "all_laid_to_args(uint64,uint64,uint64,uint64,uint64,uint64,uint64,uint64,uint64,uint64,uint64,uint64,uint64,uint64,uint64,uint64)uint64"
        )

        @pt.ABIReturnSubroutine
        def empty_return_subroutine() -> pt.Expr:
            return pt.Log(pt.Bytes("appear in both approval and clear state"))

        meth = router.add_method_handler(
            empty_return_subroutine,
            method_config=pt.MethodConfig(
                no_op=pt.CallConfig.CALL,
                opt_in=pt.CallConfig.ALL,
                clear_state=pt.CallConfig.CALL,
            ),
        )
        assert meth.method_signature() == "empty_return_subroutine()void"

        @pt.ABIReturnSubroutine
        def log_1(*, output: pt.abi.Uint64) -> pt.Expr:
            return output.set(1)

        meth = router.add_method_handler(
            log_1,
            method_config=pt.MethodConfig(
                no_op=pt.CallConfig.CALL,
                opt_in=pt.CallConfig.CALL,
                clear_state=pt.CallConfig.CALL,
            ),
        )

        assert meth.method_signature() == "log_1()uint64"

        @pt.ABIReturnSubroutine
        def log_creation(*, output: pt.abi.String) -> pt.Expr:
            return output.set("logging creation")

        meth = router.add_method_handler(
            log_creation, method_config=pt.MethodConfig(no_op=pt.CallConfig.CREATE)
        )
        assert meth.method_signature() == "log_creation()string"

        @pt.ABIReturnSubroutine
        def approve_if_odd(condition_encoding: pt.abi.Uint32) -> pt.Expr:
            return (
                pt.If(condition_encoding.get() % pt.Int(2))
                .Then(pt.Approve())
                .Else(pt.Reject())
            )

        meth = router.add_method_handler(
            approve_if_odd,
            method_config=pt.MethodConfig(
                no_op=pt.CallConfig.NEVER, clear_state=pt.CallConfig.CALL
            ),
        )
        assert meth.method_signature() == "approve_if_odd(uint32)void"

    on_completion_actions = pt.BareCallActions(
        opt_in=pt.OnCompleteAction.call_only(pt.Log(pt.Bytes("optin call"))),
        clear_state=pt.OnCompleteAction.call_only(pt.Approve()),
    )

    _router_with_oc = pt.Router(
        "ASimpleQuestionablyRobustContract", on_completion_actions
    )
    add_methods_to_router(_router_with_oc)
    (
        actual_ap_with_oc_compiled,
        actual_csp_with_oc_compiled,
        _,
    ) = _router_with_oc.compile_program(version=6)

    expected_ap_with_oc = """#pragma version 6
txn NumAppArgs
int 0
==
bnz main_l20
txna ApplicationArgs 0
method "add(uint64,uint64)uint64"
==
bnz main_l19
txna ApplicationArgs 0
method "sub(uint64,uint64)uint64"
==
bnz main_l18
txna ApplicationArgs 0
method "mul(uint64,uint64)uint64"
==
bnz main_l17
txna ApplicationArgs 0
method "div(uint64,uint64)uint64"
==
bnz main_l16
txna ApplicationArgs 0
method "mod(uint64,uint64)uint64"
==
bnz main_l15
txna ApplicationArgs 0
method "all_laid_to_args(uint64,uint64,uint64,uint64,uint64,uint64,uint64,uint64,uint64,uint64,uint64,uint64,uint64,uint64,uint64,uint64)uint64"
==
bnz main_l14
txna ApplicationArgs 0
method "empty_return_subroutine()void"
==
bnz main_l13
txna ApplicationArgs 0
method "log_1()uint64"
==
bnz main_l12
txna ApplicationArgs 0
method "log_creation()string"
==
bnz main_l11
err
main_l11:
txn OnCompletion
int NoOp
==
txn ApplicationID
int 0
==
&&
assert
callsub logcreation_8
store 67
byte 0x151f7c75
load 67
concat
log
int 1
return
main_l12:
txn OnCompletion
int NoOp
==
txn ApplicationID
int 0
!=
&&
txn OnCompletion
int OptIn
==
txn ApplicationID
int 0
!=
&&
||
assert
callsub log1_7
store 65
byte 0x151f7c75
load 65
itob
concat
log
int 1
return
main_l13:
txn OnCompletion
int NoOp
==
txn ApplicationID
int 0
!=
&&
txn OnCompletion
int OptIn
==
||
assert
callsub emptyreturnsubroutine_6
int 1
return
main_l14:
txn OnCompletion
int NoOp
==
txn ApplicationID
int 0
!=
&&
assert
txna ApplicationArgs 1
btoi
store 30
txna ApplicationArgs 2
btoi
store 31
txna ApplicationArgs 3
btoi
store 32
txna ApplicationArgs 4
btoi
store 33
txna ApplicationArgs 5
btoi
store 34
txna ApplicationArgs 6
btoi
store 35
txna ApplicationArgs 7
btoi
store 36
txna ApplicationArgs 8
btoi
store 37
txna ApplicationArgs 9
btoi
store 38
txna ApplicationArgs 10
btoi
store 39
txna ApplicationArgs 11
btoi
store 40
txna ApplicationArgs 12
btoi
store 41
txna ApplicationArgs 13
btoi
store 42
txna ApplicationArgs 14
btoi
store 43
txna ApplicationArgs 15
store 46
load 46
int 0
extract_uint64
store 44
load 46
int 8
extract_uint64
store 45
load 30
load 31
load 32
load 33
load 34
load 35
load 36
load 37
load 38
load 39
load 40
load 41
load 42
load 43
load 44
load 45
callsub alllaidtoargs_5
store 47
byte 0x151f7c75
load 47
itob
concat
log
int 1
return
main_l15:
txn OnCompletion
int NoOp
==
txn ApplicationID
int 0
!=
&&
assert
txna ApplicationArgs 1
btoi
store 24
txna ApplicationArgs 2
btoi
store 25
load 24
load 25
callsub mod_4
store 26
byte 0x151f7c75
load 26
itob
concat
log
int 1
return
main_l16:
txn OnCompletion
int NoOp
==
txn ApplicationID
int 0
!=
&&
assert
txna ApplicationArgs 1
btoi
store 18
txna ApplicationArgs 2
btoi
store 19
load 18
load 19
callsub div_3
store 20
byte 0x151f7c75
load 20
itob
concat
log
int 1
return
main_l17:
txn OnCompletion
int NoOp
==
txn ApplicationID
int 0
!=
&&
assert
txna ApplicationArgs 1
btoi
store 12
txna ApplicationArgs 2
btoi
store 13
load 12
load 13
callsub mul_2
store 14
byte 0x151f7c75
load 14
itob
concat
log
int 1
return
main_l18:
txn OnCompletion
int NoOp
==
txn ApplicationID
int 0
!=
&&
assert
txna ApplicationArgs 1
btoi
store 6
txna ApplicationArgs 2
btoi
store 7
load 6
load 7
callsub sub_1
store 8
byte 0x151f7c75
load 8
itob
concat
log
int 1
return
main_l19:
txn OnCompletion
int NoOp
==
txn ApplicationID
int 0
!=
&&
assert
txna ApplicationArgs 1
btoi
store 0
txna ApplicationArgs 2
btoi
store 1
load 0
load 1
callsub add_0
store 2
byte 0x151f7c75
load 2
itob
concat
log
int 1
return
main_l20:
txn OnCompletion
int OptIn
==
bnz main_l22
err
main_l22:
txn ApplicationID
int 0
!=
assert
byte "optin call"
log
int 1
return

// add
add_0:
store 4
store 3
load 3
load 4
+
store 5
load 5
retsub

// sub
sub_1:
store 10
store 9
load 9
load 10
-
store 11
load 11
retsub

// mul
mul_2:
store 16
store 15
load 15
load 16
*
store 17
load 17
retsub

// div
div_3:
store 22
store 21
load 21
load 22
/
store 23
load 23
retsub

// mod
mod_4:
store 28
store 27
load 27
load 28
%
store 29
load 29
retsub

// all_laid_to_args
alllaidtoargs_5:
store 63
store 62
store 61
store 60
store 59
store 58
store 57
store 56
store 55
store 54
store 53
store 52
store 51
store 50
store 49
store 48
load 48
load 49
+
load 50
+
load 51
+
load 52
+
load 53
+
load 54
+
load 55
+
load 56
+
load 57
+
load 58
+
load 59
+
load 60
+
load 61
+
load 62
+
load 63
+
store 64
load 64
retsub

// empty_return_subroutine
emptyreturnsubroutine_6:
byte "appear in both approval and clear state"
log
retsub

// log_1
log1_7:
int 1
store 66
load 66
retsub

// log_creation
logcreation_8:
byte 0x00106c6f6767696e67206372656174696f6e
store 68
load 68
retsub""".strip()

    assert expected_ap_with_oc == actual_ap_with_oc_compiled

    expected_csp_with_oc = """#pragma version 6
txn NumAppArgs
int 0
==
bnz main_l8
txna ApplicationArgs 0
method "empty_return_subroutine()void"
==
bnz main_l7
txna ApplicationArgs 0
method "log_1()uint64"
==
bnz main_l6
txna ApplicationArgs 0
method "approve_if_odd(uint32)void"
==
bnz main_l5
err
main_l5:
txna ApplicationArgs 1
int 0
extract_uint32
store 2
load 2
callsub approveifodd_2
int 1
return
main_l6:
callsub log1_1
store 1
byte 0x151f7c75
load 1
itob
concat
log
int 1
return
main_l7:
callsub emptyreturnsubroutine_0
int 1
return
main_l8:
int 1
return

// empty_return_subroutine
emptyreturnsubroutine_0:
byte "appear in both approval and clear state"
log
retsub

// log_1
log1_1:
int 1
store 0
load 0
retsub

// approve_if_odd
approveifodd_2:
store 3
load 3
int 2
%
bnz approveifodd_2_l2
int 0
return
approveifodd_2_l2:
int 1
return""".strip()
    assert expected_csp_with_oc == actual_csp_with_oc_compiled

    _router_without_oc = pt.Router("yetAnotherContractConstructedFromRouter")
    add_methods_to_router(_router_without_oc)
    (
        actual_ap_without_oc_compiled,
        actual_csp_without_oc_compiled,
        _,
    ) = _router_without_oc.compile_program(version=6)
    expected_ap_without_oc = """#pragma version 6
txna ApplicationArgs 0
method "add(uint64,uint64)uint64"
==
bnz main_l18
txna ApplicationArgs 0
method "sub(uint64,uint64)uint64"
==
bnz main_l17
txna ApplicationArgs 0
method "mul(uint64,uint64)uint64"
==
bnz main_l16
txna ApplicationArgs 0
method "div(uint64,uint64)uint64"
==
bnz main_l15
txna ApplicationArgs 0
method "mod(uint64,uint64)uint64"
==
bnz main_l14
txna ApplicationArgs 0
method "all_laid_to_args(uint64,uint64,uint64,uint64,uint64,uint64,uint64,uint64,uint64,uint64,uint64,uint64,uint64,uint64,uint64,uint64)uint64"
==
bnz main_l13
txna ApplicationArgs 0
method "empty_return_subroutine()void"
==
bnz main_l12
txna ApplicationArgs 0
method "log_1()uint64"
==
bnz main_l11
txna ApplicationArgs 0
method "log_creation()string"
==
bnz main_l10
err
main_l10:
txn OnCompletion
int NoOp
==
txn ApplicationID
int 0
==
&&
assert
callsub logcreation_8
store 67
byte 0x151f7c75
load 67
concat
log
int 1
return
main_l11:
txn OnCompletion
int NoOp
==
txn ApplicationID
int 0
!=
&&
txn OnCompletion
int OptIn
==
txn ApplicationID
int 0
!=
&&
||
assert
callsub log1_7
store 65
byte 0x151f7c75
load 65
itob
concat
log
int 1
return
main_l12:
txn OnCompletion
int NoOp
==
txn ApplicationID
int 0
!=
&&
txn OnCompletion
int OptIn
==
||
assert
callsub emptyreturnsubroutine_6
int 1
return
main_l13:
txn OnCompletion
int NoOp
==
txn ApplicationID
int 0
!=
&&
assert
txna ApplicationArgs 1
btoi
store 30
txna ApplicationArgs 2
btoi
store 31
txna ApplicationArgs 3
btoi
store 32
txna ApplicationArgs 4
btoi
store 33
txna ApplicationArgs 5
btoi
store 34
txna ApplicationArgs 6
btoi
store 35
txna ApplicationArgs 7
btoi
store 36
txna ApplicationArgs 8
btoi
store 37
txna ApplicationArgs 9
btoi
store 38
txna ApplicationArgs 10
btoi
store 39
txna ApplicationArgs 11
btoi
store 40
txna ApplicationArgs 12
btoi
store 41
txna ApplicationArgs 13
btoi
store 42
txna ApplicationArgs 14
btoi
store 43
txna ApplicationArgs 15
store 46
load 46
int 0
extract_uint64
store 44
load 46
int 8
extract_uint64
store 45
load 30
load 31
load 32
load 33
load 34
load 35
load 36
load 37
load 38
load 39
load 40
load 41
load 42
load 43
load 44
load 45
callsub alllaidtoargs_5
store 47
byte 0x151f7c75
load 47
itob
concat
log
int 1
return
main_l14:
txn OnCompletion
int NoOp
==
txn ApplicationID
int 0
!=
&&
assert
txna ApplicationArgs 1
btoi
store 24
txna ApplicationArgs 2
btoi
store 25
load 24
load 25
callsub mod_4
store 26
byte 0x151f7c75
load 26
itob
concat
log
int 1
return
main_l15:
txn OnCompletion
int NoOp
==
txn ApplicationID
int 0
!=
&&
assert
txna ApplicationArgs 1
btoi
store 18
txna ApplicationArgs 2
btoi
store 19
load 18
load 19
callsub div_3
store 20
byte 0x151f7c75
load 20
itob
concat
log
int 1
return
main_l16:
txn OnCompletion
int NoOp
==
txn ApplicationID
int 0
!=
&&
assert
txna ApplicationArgs 1
btoi
store 12
txna ApplicationArgs 2
btoi
store 13
load 12
load 13
callsub mul_2
store 14
byte 0x151f7c75
load 14
itob
concat
log
int 1
return
main_l17:
txn OnCompletion
int NoOp
==
txn ApplicationID
int 0
!=
&&
assert
txna ApplicationArgs 1
btoi
store 6
txna ApplicationArgs 2
btoi
store 7
load 6
load 7
callsub sub_1
store 8
byte 0x151f7c75
load 8
itob
concat
log
int 1
return
main_l18:
txn OnCompletion
int NoOp
==
txn ApplicationID
int 0
!=
&&
assert
txna ApplicationArgs 1
btoi
store 0
txna ApplicationArgs 2
btoi
store 1
load 0
load 1
callsub add_0
store 2
byte 0x151f7c75
load 2
itob
concat
log
int 1
return

// add
add_0:
store 4
store 3
load 3
load 4
+
store 5
load 5
retsub

// sub
sub_1:
store 10
store 9
load 9
load 10
-
store 11
load 11
retsub

// mul
mul_2:
store 16
store 15
load 15
load 16
*
store 17
load 17
retsub

// div
div_3:
store 22
store 21
load 21
load 22
/
store 23
load 23
retsub

// mod
mod_4:
store 28
store 27
load 27
load 28
%
store 29
load 29
retsub

// all_laid_to_args
alllaidtoargs_5:
store 63
store 62
store 61
store 60
store 59
store 58
store 57
store 56
store 55
store 54
store 53
store 52
store 51
store 50
store 49
store 48
load 48
load 49
+
load 50
+
load 51
+
load 52
+
load 53
+
load 54
+
load 55
+
load 56
+
load 57
+
load 58
+
load 59
+
load 60
+
load 61
+
load 62
+
load 63
+
store 64
load 64
retsub

// empty_return_subroutine
emptyreturnsubroutine_6:
byte "appear in both approval and clear state"
log
retsub

// log_1
log1_7:
int 1
store 66
load 66
retsub

// log_creation
logcreation_8:
byte 0x00106c6f6767696e67206372656174696f6e
store 68
load 68
retsub""".strip()
    assert actual_ap_without_oc_compiled == expected_ap_without_oc

    expected_csp_without_oc = """#pragma version 6
txna ApplicationArgs 0
method "empty_return_subroutine()void"
==
bnz main_l6
txna ApplicationArgs 0
method "log_1()uint64"
==
bnz main_l5
txna ApplicationArgs 0
method "approve_if_odd(uint32)void"
==
bnz main_l4
err
main_l4:
txna ApplicationArgs 1
int 0
extract_uint32
store 2
load 2
callsub approveifodd_2
int 1
return
main_l5:
callsub log1_1
store 1
byte 0x151f7c75
load 1
itob
concat
log
int 1
return
main_l6:
callsub emptyreturnsubroutine_0
int 1
return

// empty_return_subroutine
emptyreturnsubroutine_0:
byte "appear in both approval and clear state"
log
retsub

// log_1
log1_1:
int 1
store 0
load 0
retsub

// approve_if_odd
approveifodd_2:
store 3
load 3
int 2
%
bnz approveifodd_2_l2
int 0
return
approveifodd_2_l2:
int 1
return""".strip()
    assert actual_csp_without_oc_compiled == expected_csp_without_oc



from itertools import product
from os import environ
from pathlib import Path
import pytest

from pyteal import compileTeal, Mode

import examples.signature.factorizer_game as factorizer

from tests.blackbox import algod_with_assertion
from graviton.blackbox import (
    DryRunExecutor as Executor,
    DryRunInspector as Inspector,
    DryRunProperty as DRProp,
)
from graviton.invariant import Invariant

REPORTS_DIR = Path.cwd() / "tests" / "integration" / "reports"
ALGOD = algod_with_assertion()

DEFAULT = {
    "A": 3,
    "P": 5,  # 13
    "Q": 7,  # 13
    "M": 5,  # 10
    "N": 5,  # 10
}


def get_param_bounds():
    """
    Allow setting the bounds either from the environment via something like:

    % A=3 P=13 Q=13 M=10 N=10 pytest tests/integration/pure_logicsig_test.py::test_many_factorizer_games

    OR - when any of the above is missing, replace with the default version
    """
    vars = []
    for var in ("A", "P", "Q", "M", "N"):
        val = environ.get(var)
        if val is None:
            val = DEFAULT[var]
        vars.append(int(val))
    return vars


def get_factorizer_param_sequence():
    A, P, Q, M, N = get_param_bounds()
    return [(a, p, q, M, N) for a in range(A) for p in range(P) for q in range(Q)]


def inputs_for_coefficients(a, p, q, M, N):
    # TODO: this should really be focused around the roots p and q
    return product(range(M), range(N))


def factorizer_game_check(a: int, p: int, q: int, M: int, N: int):
    ae = None
    if a <= 0 or p < 0 or q <= p:
        with pytest.raises(AssertionError) as ae:
            factorizer.logicsig(a, p, q),

    if ae:
        return

    compiled = compileTeal(
        factorizer.logicsig(a, p, q),
        version=6,
        mode=Mode.Signature,
        assembleConstants=True,
    )
    inputs = list(inputs_for_coefficients(a, p, q, M, N))
    N = len(inputs)

    def poly(x):
        return abs(a * x**2 - a * (p + q) * x + a * p * q)

    def naive_prize(x, y):
        return 1_000_000 * max(10 - (sum(map(poly, (x, y))) + 1) // 2, 0)

    def payment_amount(x, y):
        return 0 if x == y else naive_prize(x, y)

    amts = list(map(lambda args: payment_amount(*args), inputs))

    inspectors, txns = [], []
    for args, amt in zip(inputs, amts):
        txn = {"amt": amt}
        txns.append(txn)
        inspectors.append(Executor.dryrun_logicsig(ALGOD, compiled, args, **txn))

    print(
        f"generating a report for (a,p,q) = {a,p,q} with {M, N} dry-run calls and spreadsheet rows"
    )
    filebase = f"factorizer_game_{a}_{p}_{q}"

    reports_dir = REPORTS_DIR / "pure_logicsig"
    reports_dir.mkdir(parents=True, exist_ok=True)
    csvpath = reports_dir / f"{filebase}.csv"
    with open(csvpath, "w") as f:
        f.write(Inspector.csv_report(inputs, inspectors, txns=txns))

    print(f"validating passing_invariant for (a,p,q) = {a,p,q} over {N} dry-run calls")
    passing_invariant = Invariant(
        lambda args: bool(payment_amount(*args)),
        name=f"passing invariant for coeffs {a, p, q}",
    )
    passing_invariant.validates(DRProp.passed, inputs, inspectors)

    print(
        f"validate procedurally that payment amount as expected for (a,p,q) = {a,p,q} over {M, N} dry-rundry-run calls"
    )

    for args, inspector in zip(inputs, inspectors):
        x, y = args
        eprize = naive_prize(x, y)
        final_scratches = inspector.final_scratch().values()
        assert eprize == 0 or eprize in final_scratches, inspector.report(
            args,
            f"(a, p, q, x, y) = {a, p, q, x, y}. final scratch slots expected to contain {eprize} v. actual={final_scratches}",
        )


@pytest.mark.parametrize("a, p, q, M, N", get_factorizer_param_sequence())
def test_many_factorizer_games(a: int, p: int, q: int, M: int, N: int):
    factorizer_game_check(a, p, q, M, N)



import pyteal as pt


def test_constructor():
    block1 = pt.TealSimpleBlock([])
    assert block1.ops == []
    assert block1.nextBlock is None

    block2 = pt.TealSimpleBlock([pt.TealOp(None, pt.Op.int, 1)])
    assert block2.ops == [pt.TealOp(None, pt.Op.int, 1)]
    assert block2.nextBlock is None


def test_next_block():
    block = pt.TealSimpleBlock([])
    block.setNextBlock(pt.TealSimpleBlock([pt.TealOp(None, pt.Op.substring3)]))
    assert block.nextBlock == pt.TealSimpleBlock([pt.TealOp(None, pt.Op.substring3)])


def test_outgoing():
    emptyBlock = pt.TealSimpleBlock([])
    assert emptyBlock.getOutgoing() == []

    block = pt.TealSimpleBlock([])
    block.setNextBlock(pt.TealSimpleBlock([pt.TealOp(None, pt.Op.byte, '"nextBlock"')]))
    assert block.getOutgoing() == [
        pt.TealSimpleBlock([pt.TealOp(None, pt.Op.byte, '"nextBlock"')])
    ]



from typing import Any, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from pyteal.ast import Expr


class TealInternalError(Exception):
    def __init__(self, message: str) -> None:
        self.message = message

    def __str__(self):
        return self.message


TealInternalError.__module__ = "pyteal"


class TealTypeError(Exception):
    def __init__(self, actual, expected) -> None:
        self.message = "{} while expected {} ".format(actual, expected)

    def __str__(self) -> str:
        return self.message


TealTypeError.__module__ = "pyteal"


class TealInputError(Exception):
    def __init__(self, msg: str) -> None:
        self.message = msg

    def __str__(self) -> str:
        return self.message

    def __eq__(self, other: Any) -> bool:
        return type(other) is TealInputError and self.message == other.message


TealInputError.__module__ = "pyteal"


class TealCompileError(Exception):
    def __init__(self, msg: str, sourceExpr: Optional["Expr"]) -> None:
        self.msg = msg
        self.sourceExpr = sourceExpr

    def __str__(self) -> str:
        if self.sourceExpr is None:
            return self.msg
        trace = self.sourceExpr.getDefinitionTrace()
        return (
            self.msg
            + "\nTraceback of origin expression (most recent call last):\n"
            + "".join(trace)
        )

    def __eq__(self, other) -> bool:
        if not isinstance(other, TealCompileError):
            return False
        return self.msg == other.msg and self.sourceExpr is other.sourceExpr


TealCompileError.__module__ = "pyteal"


class TealPragmaError(Exception):
    def __init__(self, message: str) -> None:
        self.message = message

    def __str__(self):
        return self.message


TealPragmaError.__module__ = "pyteal"


def verifyProgramVersion(minVersion: int, version: int, msg: str):
    if minVersion > version:
        msg = "{}. Minimum version needed is {}, but current version being compiled is {}".format(
            msg, minVersion, version
        )
        raise TealInputError(msg)


def verifyFieldVersion(fieldName: str, fieldMinVersion: int, version: int):
    verifyProgramVersion(
        fieldMinVersion,
        version,
        "Program version too low to use field {}".format(fieldName),
    )



## File generated from scripts/generate_init.py.
## DO NOT EDIT DIRECTLY

from pyteal.ast import *
from pyteal.ast import __all__ as ast_all
from pyteal.pragma import pragma
from pyteal.ir import *
from pyteal.ir import __all__ as ir_all
from pyteal.compiler import (
    MAX_TEAL_VERSION,
    MIN_TEAL_VERSION,
    DEFAULT_TEAL_VERSION,
    MAX_PROGRAM_VERSION,
    MIN_PROGRAM_VERSION,
    DEFAULT_PROGRAM_VERSION,
    CompileOptions,
    compileTeal,
    OptimizeOptions,
)
from pyteal.types import TealType
from pyteal.errors import (
    TealInternalError,
    TealTypeError,
    TealInputError,
    TealCompileError,
    TealPragmaError,
)
from pyteal.config import (
    MAX_GROUP_SIZE,
    NUM_SLOTS,
    RETURN_HASH_PREFIX,
    METHOD_ARG_NUM_CUTOFF,
)

__all__ = [
    "ABIReturnSubroutine",
    "AccountParam",
    "AccountParamObject",
    "Add",
    "Addr",
    "And",
    "App",
    "AppField",
    "AppParam",
    "AppParamObject",
    "Approve",
    "Arg",
    "Array",
    "Assert",
    "AssetHolding",
    "AssetHoldingObject",
    "AssetParam",
    "AssetParamObject",
    "Balance",
    "BareCallActions",
    "Base64Decode",
    "BinaryExpr",
    "BitLen",
    "BitwiseAnd",
    "BitwiseNot",
    "BitwiseOr",
    "BitwiseXor",
    "Block",
    "Break",
    "Btoi",
    "Bytes",
    "BytesAdd",
    "BytesAnd",
    "BytesDiv",
    "BytesEq",
    "BytesGe",
    "BytesGt",
    "BytesLe",
    "BytesLt",
    "BytesMinus",
    "BytesMod",
    "BytesMul",
    "BytesNeq",
    "BytesNot",
    "BytesOr",
    "BytesSqrt",
    "BytesXor",
    "BytesZero",
    "CallConfig",
    "Comment",
    "CompileOptions",
    "Concat",
    "Cond",
    "Continue",
    "DEFAULT_PROGRAM_VERSION",
    "DEFAULT_TEAL_VERSION",
    "Div",
    "Divw",
    "DynamicScratchVar",
    "EcdsaCurve",
    "EcdsaDecompress",
    "EcdsaRecover",
    "EcdsaVerify",
    "Ed25519Verify",
    "Ed25519Verify_Bare",
    "EnumInt",
    "Eq",
    "Err",
    "Exp",
    "Expr",
    "Extract",
    "ExtractUint16",
    "ExtractUint32",
    "ExtractUint64",
    "For",
    "Ge",
    "GeneratedID",
    "GetBit",
    "GetByte",
    "Gitxn",
    "GitxnExpr",
    "GitxnaExpr",
    "Global",
    "GlobalField",
    "Gt",
    "Gtxn",
    "GtxnExpr",
    "GtxnaExpr",
    "If",
    "ImportScratchValue",
    "InnerTxn",
    "InnerTxnAction",
    "InnerTxnBuilder",
    "InnerTxnGroup",
    "Int",
    "Itob",
    "JsonRef",
    "Keccak256",
    "LabelReference",
    "Le",
    "LeafExpr",
    "Len",
    "Log",
    "Lt",
    "MAX_GROUP_SIZE",
    "MAX_PROGRAM_VERSION",
    "MAX_TEAL_VERSION",
    "METHOD_ARG_NUM_CUTOFF",
    "MIN_PROGRAM_VERSION",
    "MIN_TEAL_VERSION",
    "MaybeValue",
    "MethodConfig",
    "MethodSignature",
    "MinBalance",
    "Minus",
    "Mod",
    "Mode",
    "Mul",
    "MultiValue",
    "NUM_SLOTS",
    "NaryExpr",
    "Neq",
    "Nonce",
    "Not",
    "OnComplete",
    "OnCompleteAction",
    "Op",
    "OpUp",
    "OpUpFeeSource",
    "OpUpMode",
    "OptimizeOptions",
    "Or",
    "Pop",
    "Pragma",
    "RETURN_HASH_PREFIX",
    "Reject",
    "Replace",
    "Return",
    "Router",
    "ScratchIndex",
    "ScratchLoad",
    "ScratchSlot",
    "ScratchStackStore",
    "ScratchStore",
    "ScratchVar",
    "Seq",
    "SetBit",
    "SetByte",
    "Sha256",
    "Sha3_256",
    "Sha512_256",
    "ShiftLeft",
    "ShiftRight",
    "Sqrt",
    "Subroutine",
    "SubroutineCall",
    "SubroutineDeclaration",
    "SubroutineDefinition",
    "SubroutineFnWrapper",
    "Substring",
    "Suffix",
    "TealBlock",
    "TealCompileError",
    "TealComponent",
    "TealConditionalBlock",
    "TealInputError",
    "TealInternalError",
    "TealLabel",
    "TealOp",
    "TealPragmaError",
    "TealSimpleBlock",
    "TealType",
    "TealTypeError",
    "Tmpl",
    "Txn",
    "TxnArray",
    "TxnExpr",
    "TxnField",
    "TxnGroup",
    "TxnObject",
    "TxnType",
    "TxnaExpr",
    "UnaryExpr",
    "VrfVerify",
    "While",
    "WideRatio",
    "abi",
    "compileTeal",
    "pragma",
]



from typing import TYPE_CHECKING, Optional
from pyteal.ast.seq import _use_seq_if_multiple

from pyteal.errors import TealCompileError
from pyteal.types import TealType, require_type
from pyteal.ir import TealSimpleBlock, TealConditionalBlock
from pyteal.ast.expr import Expr

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class While(Expr):
    """While expression."""

    def __init__(self, cond: Expr) -> None:
        """Create a new While expression.

        When this While expression is executed, the condition will be evaluated, and if it produces a
        true value, doBlock will be executed and return to the start of the expression execution.
        Otherwise, no branch will be executed.

        Args:
            cond: The condition to check. Must evaluate to uint64.

        Example:
            .. code-block:: python

                i = ScratchVar()
                i.store(Int(0))
                While(i.load() < pt.Int(2))
                    .Do(Pop(Int(1)), i.store(i.load() + Int(1)))
        """
        super().__init__()
        require_type(cond, TealType.uint64)

        self.cond = cond
        self.doBlock: Optional[Expr] = None

    def __teal__(self, options: "CompileOptions"):
        if self.doBlock is None:
            raise TealCompileError("While expression must have a doBlock", self)

        options.enterLoop()

        condStart, condEnd = self.cond.__teal__(options)
        doStart, doEnd = self.doBlock.__teal__(options)
        end = TealSimpleBlock([])

        doEnd.setNextBlock(condStart)

        branchBlock = TealConditionalBlock([])
        branchBlock.setTrueBlock(doStart)
        branchBlock.setFalseBlock(end)

        condEnd.setNextBlock(branchBlock)

        breakBlocks, continueBlocks = options.exitLoop()

        for block in breakBlocks:
            block.setNextBlock(end)

        for block in continueBlocks:
            block.setNextBlock(condStart)

        return condStart, end

    def __str__(self):
        if self.doBlock is None:
            raise TealCompileError("While expression must have a doBlock", self)

        return "(While {} {})".format(self.cond, self.doBlock)

    def type_of(self):
        if self.doBlock is None:
            raise TealCompileError("While expression must have a doBlock", self)
        return TealType.none

    def has_return(self):
        return False

    def Do(self, doBlock: Expr, *do_block_multi: Expr):
        if self.doBlock is not None:
            raise TealCompileError("While expression already has a doBlock", self)

        doBlock = _use_seq_if_multiple(doBlock, *do_block_multi)

        require_type(doBlock, TealType.none)
        self.doBlock = doBlock
        return self


While.__module__ = "pyteal"



import pytest

import pyteal as pt

avm2Options = pt.CompileOptions(version=2)
avm3Options = pt.CompileOptions(version=3)
avm4Options = pt.CompileOptions(version=4)
avm5Options = pt.CompileOptions(version=5)


def test_add():
    args = [pt.Int(2), pt.Int(3)]
    expr = pt.Add(args[0], args[1])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 2),
            pt.TealOp(args[1], pt.Op.int, 3),
            pt.TealOp(expr, pt.Op.add),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_add_overload():
    args = [pt.Int(2), pt.Int(3), pt.Int(4)]
    expr = args[0] + args[1] + args[2]
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 2),
            pt.TealOp(args[1], pt.Op.int, 3),
            pt.TealOp(None, pt.Op.add),
            pt.TealOp(args[2], pt.Op.int, 4),
            pt.TealOp(None, pt.Op.add),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_add_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.Add(pt.Int(2), pt.Txn.receiver())

    with pytest.raises(pt.TealTypeError):
        pt.Add(pt.Txn.sender(), pt.Int(2))


def test_minus():
    args = [pt.Int(5), pt.Int(6)]
    expr = pt.Minus(args[0], args[1])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 5),
            pt.TealOp(args[1], pt.Op.int, 6),
            pt.TealOp(expr, pt.Op.minus),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_minus_overload():
    args = [pt.Int(10), pt.Int(1), pt.Int(2)]
    expr = args[0] - args[1] - args[2]
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 10),
            pt.TealOp(args[1], pt.Op.int, 1),
            pt.TealOp(None, pt.Op.minus),
            pt.TealOp(args[2], pt.Op.int, 2),
            pt.TealOp(None, pt.Op.minus),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_minus_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.Minus(pt.Int(2), pt.Txn.receiver())

    with pytest.raises(pt.TealTypeError):
        pt.Minus(pt.Txn.sender(), pt.Int(2))


def test_mul():
    args = [pt.Int(3), pt.Int(8)]
    expr = pt.Mul(args[0], args[1])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 3),
            pt.TealOp(args[1], pt.Op.int, 8),
            pt.TealOp(expr, pt.Op.mul),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_mul_overload():
    args = [pt.Int(3), pt.Int(8), pt.Int(10)]
    expr = args[0] * args[1] * args[2]
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 3),
            pt.TealOp(args[1], pt.Op.int, 8),
            pt.TealOp(None, pt.Op.mul),
            pt.TealOp(args[2], pt.Op.int, 10),
            pt.TealOp(None, pt.Op.mul),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_mul_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.Mul(pt.Int(2), pt.Txn.receiver())

    with pytest.raises(pt.TealTypeError):
        pt.Mul(pt.Txn.sender(), pt.Int(2))


def test_div():
    args = [pt.Int(9), pt.Int(3)]
    expr = pt.Div(args[0], args[1])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 9),
            pt.TealOp(args[1], pt.Op.int, 3),
            pt.TealOp(expr, pt.Op.div),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_div_overload():
    args = [pt.Int(9), pt.Int(3), pt.Int(3)]
    expr = args[0] / args[1] / args[2]
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 9),
            pt.TealOp(args[1], pt.Op.int, 3),
            pt.TealOp(None, pt.Op.div),
            pt.TealOp(args[2], pt.Op.int, 3),
            pt.TealOp(None, pt.Op.div),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_div_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.Div(pt.Int(2), pt.Txn.receiver())

    with pytest.raises(pt.TealTypeError):
        pt.Div(pt.Txn.sender(), pt.Int(2))


def test_mod():
    args = [pt.Int(10), pt.Int(9)]
    expr = pt.Mod(args[0], args[1])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 10),
            pt.TealOp(args[1], pt.Op.int, 9),
            pt.TealOp(expr, pt.Op.mod),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_mod_overload():
    args = [pt.Int(10), pt.Int(9), pt.Int(100)]
    expr = args[0] % args[1] % args[2]
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 10),
            pt.TealOp(args[1], pt.Op.int, 9),
            pt.TealOp(None, pt.Op.mod),
            pt.TealOp(args[2], pt.Op.int, 100),
            pt.TealOp(None, pt.Op.mod),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_mod_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.Mod(pt.Txn.receiver(), pt.Int(2))

    with pytest.raises(pt.TealTypeError):
        pt.Mod(pt.Int(2), pt.Txn.sender())


def test_exp():
    args = [pt.Int(2), pt.Int(9)]
    expr = pt.Exp(args[0], args[1])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 2),
            pt.TealOp(args[1], pt.Op.int, 9),
            pt.TealOp(expr, pt.Op.exp),
        ]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm3Options)


def test_exp_overload():
    args = [pt.Int(2), pt.Int(3), pt.Int(1)]
    # this is equivalent to args[0] ** (args[1] ** args[2])
    expr = args[0] ** args[1] ** args[2]
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 2),
            pt.TealOp(args[1], pt.Op.int, 3),
            pt.TealOp(args[2], pt.Op.int, 1),
            pt.TealOp(None, pt.Op.exp),
            pt.TealOp(None, pt.Op.exp),
        ]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm3Options)


def test_exp_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.Exp(pt.Txn.receiver(), pt.Int(2))

    with pytest.raises(pt.TealTypeError):
        pt.Exp(pt.Int(2), pt.Txn.sender())


def test_arithmetic():
    args = [pt.Int(2), pt.Int(3), pt.Int(5), pt.Int(6), pt.Int(8), pt.Int(9)]
    v = ((args[0] + args[1]) / ((args[2] - args[3]) * args[4])) % args[5]
    assert v.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 2),
            pt.TealOp(args[1], pt.Op.int, 3),
            pt.TealOp(None, pt.Op.add),
            pt.TealOp(args[2], pt.Op.int, 5),
            pt.TealOp(args[3], pt.Op.int, 6),
            pt.TealOp(None, pt.Op.minus),
            pt.TealOp(args[4], pt.Op.int, 8),
            pt.TealOp(None, pt.Op.mul),
            pt.TealOp(None, pt.Op.div),
            pt.TealOp(args[5], pt.Op.int, 9),
            pt.TealOp(None, pt.Op.mod),
        ]
    )

    actual, _ = v.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_bitwise_and():
    args = [pt.Int(1), pt.Int(2)]
    expr = pt.BitwiseAnd(args[0], args[1])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 1),
            pt.TealOp(args[1], pt.Op.int, 2),
            pt.TealOp(expr, pt.Op.bitwise_and),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_bitwise_and_overload():
    args = [pt.Int(1), pt.Int(2), pt.Int(4)]
    expr = args[0] & args[1] & args[2]
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 1),
            pt.TealOp(args[1], pt.Op.int, 2),
            pt.TealOp(None, pt.Op.bitwise_and),
            pt.TealOp(args[2], pt.Op.int, 4),
            pt.TealOp(None, pt.Op.bitwise_and),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_bitwise_and_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.BitwiseAnd(pt.Int(2), pt.Txn.receiver())

    with pytest.raises(pt.TealTypeError):
        pt.BitwiseAnd(pt.Txn.sender(), pt.Int(2))


def test_bitwise_or():
    args = [pt.Int(1), pt.Int(2)]
    expr = pt.BitwiseOr(args[0], args[1])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 1),
            pt.TealOp(args[1], pt.Op.int, 2),
            pt.TealOp(expr, pt.Op.bitwise_or),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_bitwise_or_overload():
    args = [pt.Int(1), pt.Int(2), pt.Int(4)]
    expr = args[0] | args[1] | args[2]
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 1),
            pt.TealOp(args[1], pt.Op.int, 2),
            pt.TealOp(None, pt.Op.bitwise_or),
            pt.TealOp(args[2], pt.Op.int, 4),
            pt.TealOp(None, pt.Op.bitwise_or),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_bitwise_or_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.BitwiseOr(pt.Int(2), pt.Txn.receiver())

    with pytest.raises(pt.TealTypeError):
        pt.BitwiseOr(pt.Txn.sender(), pt.Int(2))


def test_bitwise_xor():
    args = [pt.Int(1), pt.Int(3)]
    expr = pt.BitwiseXor(args[0], args[1])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 1),
            pt.TealOp(args[1], pt.Op.int, 3),
            pt.TealOp(expr, pt.Op.bitwise_xor),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_bitwise_xor_overload():
    args = [pt.Int(1), pt.Int(3), pt.Int(5)]
    expr = args[0] ^ args[1] ^ args[2]
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 1),
            pt.TealOp(args[1], pt.Op.int, 3),
            pt.TealOp(None, pt.Op.bitwise_xor),
            pt.TealOp(args[2], pt.Op.int, 5),
            pt.TealOp(None, pt.Op.bitwise_xor),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_bitwise_xor_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.BitwiseXor(pt.Int(2), pt.Txn.receiver())

    with pytest.raises(pt.TealTypeError):
        pt.BitwiseXor(pt.Txn.sender(), pt.Int(2))


def test_shift_left():
    args = [pt.Int(5), pt.Int(1)]
    expr = pt.ShiftLeft(args[0], args[1])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 5),
            pt.TealOp(args[1], pt.Op.int, 1),
            pt.TealOp(expr, pt.Op.shl),
        ]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm3Options)


def test_shift_left_overload():
    args = [pt.Int(5), pt.Int(1), pt.Int(2)]
    expr = args[0] << args[1] << args[2]
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 5),
            pt.TealOp(args[1], pt.Op.int, 1),
            pt.TealOp(None, pt.Op.shl),
            pt.TealOp(args[2], pt.Op.int, 2),
            pt.TealOp(None, pt.Op.shl),
        ]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm3Options)


def test_shift_left_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.ShiftLeft(pt.Int(2), pt.Txn.receiver())

    with pytest.raises(pt.TealTypeError):
        pt.ShiftLeft(pt.Txn.sender(), pt.Int(2))


def test_shift_right():
    args = [pt.Int(5), pt.Int(1)]
    expr = pt.ShiftRight(args[0], args[1])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 5),
            pt.TealOp(args[1], pt.Op.int, 1),
            pt.TealOp(expr, pt.Op.shr),
        ]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm3Options)


def test_shift_right_overload():
    args = [pt.Int(5), pt.Int(1), pt.Int(2)]
    expr = args[0] >> args[1] >> args[2]
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 5),
            pt.TealOp(args[1], pt.Op.int, 1),
            pt.TealOp(None, pt.Op.shr),
            pt.TealOp(args[2], pt.Op.int, 2),
            pt.TealOp(None, pt.Op.shr),
        ]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm3Options)


def test_shift_right_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.ShiftRight(pt.Int(2), pt.Txn.receiver())

    with pytest.raises(pt.TealTypeError):
        pt.ShiftRight(pt.Txn.sender(), pt.Int(2))


def test_eq():
    args_int = [pt.Int(2), pt.Int(3)]
    expr_int = pt.Eq(args_int[0], args_int[1])
    assert expr_int.type_of() == pt.TealType.uint64

    expected_int = pt.TealSimpleBlock(
        [
            pt.TealOp(args_int[0], pt.Op.int, 2),
            pt.TealOp(args_int[1], pt.Op.int, 3),
            pt.TealOp(expr_int, pt.Op.eq),
        ]
    )

    actual_int, _ = expr_int.__teal__(avm2Options)
    actual_int.addIncoming()
    actual_int = pt.TealBlock.NormalizeBlocks(actual_int)

    assert actual_int == expected_int

    args_bytes = [pt.Txn.receiver(), pt.Txn.sender()]
    expr_bytes = pt.Eq(args_bytes[0], args_bytes[1])
    assert expr_bytes.type_of() == pt.TealType.uint64

    expected_bytes = pt.TealSimpleBlock(
        [
            pt.TealOp(args_bytes[0], pt.Op.txn, "Receiver"),
            pt.TealOp(args_bytes[1], pt.Op.txn, "Sender"),
            pt.TealOp(expr_bytes, pt.Op.eq),
        ]
    )

    actual_bytes, _ = expr_bytes.__teal__(avm2Options)
    actual_bytes.addIncoming()
    actual_bytes = pt.TealBlock.NormalizeBlocks(actual_bytes)

    assert actual_bytes == expected_bytes


def test_eq_overload():
    args_int = [pt.Int(2), pt.Int(3)]
    expr_int = args_int[0] == args_int[1]
    assert expr_int.type_of() == pt.TealType.uint64

    expected_int = pt.TealSimpleBlock(
        [
            pt.TealOp(args_int[0], pt.Op.int, 2),
            pt.TealOp(args_int[1], pt.Op.int, 3),
            pt.TealOp(expr_int, pt.Op.eq),
        ]
    )

    actual_int, _ = expr_int.__teal__(avm2Options)
    actual_int.addIncoming()
    actual_int = pt.TealBlock.NormalizeBlocks(actual_int)

    assert actual_int == expected_int

    args_bytes = [pt.Txn.receiver(), pt.Txn.sender()]
    expr_bytes = args_bytes[0] == args_bytes[1]
    assert expr_bytes.type_of() == pt.TealType.uint64

    expected_bytes = pt.TealSimpleBlock(
        [
            pt.TealOp(args_bytes[0], pt.Op.txn, "Receiver"),
            pt.TealOp(args_bytes[1], pt.Op.txn, "Sender"),
            pt.TealOp(expr_bytes, pt.Op.eq),
        ]
    )

    actual_bytes, _ = expr_bytes.__teal__(avm2Options)
    actual_bytes.addIncoming()
    actual_bytes = pt.TealBlock.NormalizeBlocks(actual_bytes)

    assert actual_bytes == expected_bytes


def test_eq_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.Eq(pt.Txn.fee(), pt.Txn.receiver())

    with pytest.raises(pt.TealTypeError):
        pt.Eq(pt.Txn.sender(), pt.Int(7))


def test_neq():
    args_int = [pt.Int(2), pt.Int(3)]
    expr_int = pt.Neq(args_int[0], args_int[1])
    assert expr_int.type_of() == pt.TealType.uint64

    expected_int = pt.TealSimpleBlock(
        [
            pt.TealOp(args_int[0], pt.Op.int, 2),
            pt.TealOp(args_int[1], pt.Op.int, 3),
            pt.TealOp(expr_int, pt.Op.neq),
        ]
    )

    actual_int, _ = expr_int.__teal__(avm2Options)
    actual_int.addIncoming()
    actual_int = pt.TealBlock.NormalizeBlocks(actual_int)

    assert actual_int == expected_int

    args_bytes = [pt.Txn.receiver(), pt.Txn.sender()]
    expr_bytes = pt.Neq(args_bytes[0], args_bytes[1])
    assert expr_bytes.type_of() == pt.TealType.uint64

    expected_bytes = pt.TealSimpleBlock(
        [
            pt.TealOp(args_bytes[0], pt.Op.txn, "Receiver"),
            pt.TealOp(args_bytes[1], pt.Op.txn, "Sender"),
            pt.TealOp(expr_bytes, pt.Op.neq),
        ]
    )

    actual_bytes, _ = expr_bytes.__teal__(avm2Options)
    actual_bytes.addIncoming()
    actual_bytes = pt.TealBlock.NormalizeBlocks(actual_bytes)

    assert actual_bytes == expected_bytes


def test_neq_overload():
    args_int = [pt.Int(2), pt.Int(3)]
    expr_int = args_int[0] != args_int[1]
    assert expr_int.type_of() == pt.TealType.uint64

    expected_int = pt.TealSimpleBlock(
        [
            pt.TealOp(args_int[0], pt.Op.int, 2),
            pt.TealOp(args_int[1], pt.Op.int, 3),
            pt.TealOp(expr_int, pt.Op.neq),
        ]
    )

    actual_int, _ = expr_int.__teal__(avm2Options)
    actual_int.addIncoming()
    actual_int = pt.TealBlock.NormalizeBlocks(actual_int)

    assert actual_int == expected_int

    args_bytes = [pt.Txn.receiver(), pt.Txn.sender()]
    expr_bytes = args_bytes[0] != args_bytes[1]
    assert expr_bytes.type_of() == pt.TealType.uint64

    expected_bytes = pt.TealSimpleBlock(
        [
            pt.TealOp(args_bytes[0], pt.Op.txn, "Receiver"),
            pt.TealOp(args_bytes[1], pt.Op.txn, "Sender"),
            pt.TealOp(expr_bytes, pt.Op.neq),
        ]
    )

    actual_bytes, _ = expr_bytes.__teal__(avm2Options)
    actual_bytes.addIncoming()
    actual_bytes = pt.TealBlock.NormalizeBlocks(actual_bytes)

    assert actual_bytes == expected_bytes


def test_neq_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.Neq(pt.Txn.fee(), pt.Txn.receiver())

    with pytest.raises(pt.TealTypeError):
        pt.Neq(pt.Txn.sender(), pt.Int(7))


def test_lt():
    args = [pt.Int(2), pt.Int(3)]
    expr = pt.Lt(args[0], args[1])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 2),
            pt.TealOp(args[1], pt.Op.int, 3),
            pt.TealOp(expr, pt.Op.lt),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_lt_overload():
    args = [pt.Int(2), pt.Int(3)]
    expr = args[0] < args[1]
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 2),
            pt.TealOp(args[1], pt.Op.int, 3),
            pt.TealOp(expr, pt.Op.lt),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_lt_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.Lt(pt.Int(7), pt.Txn.receiver())

    with pytest.raises(pt.TealTypeError):
        pt.Lt(pt.Txn.sender(), pt.Int(7))


def test_le():
    args = [pt.Int(1), pt.Int(2)]
    expr = pt.Le(args[0], args[1])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 1),
            pt.TealOp(args[1], pt.Op.int, 2),
            pt.TealOp(expr, pt.Op.le),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_le_overload():
    args = [pt.Int(1), pt.Int(2)]
    expr = args[0] <= args[1]
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 1),
            pt.TealOp(args[1], pt.Op.int, 2),
            pt.TealOp(expr, pt.Op.le),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_le_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.Le(pt.Int(1), pt.Txn.receiver())

    with pytest.raises(pt.TealTypeError):
        pt.Le(pt.Txn.sender(), pt.Int(1))


def test_gt():
    args = [pt.Int(2), pt.Int(3)]
    expr = pt.Gt(args[0], args[1])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 2),
            pt.TealOp(args[1], pt.Op.int, 3),
            pt.TealOp(expr, pt.Op.gt),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_gt_overload():
    args = [pt.Int(2), pt.Int(3)]
    expr = args[0] > args[1]
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 2),
            pt.TealOp(args[1], pt.Op.int, 3),
            pt.TealOp(expr, pt.Op.gt),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_gt_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.Gt(pt.Int(1), pt.Txn.receiver())

    with pytest.raises(pt.TealTypeError):
        pt.Gt(pt.Txn.receiver(), pt.Int(1))


def test_ge():
    args = [pt.Int(1), pt.Int(10)]
    expr = pt.Ge(args[0], args[1])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 1),
            pt.TealOp(args[1], pt.Op.int, 10),
            pt.TealOp(expr, pt.Op.ge),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_ge_overload():
    args = [pt.Int(1), pt.Int(10)]
    expr = args[0] >= args[1]
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 1),
            pt.TealOp(args[1], pt.Op.int, 10),
            pt.TealOp(expr, pt.Op.ge),
        ]
    )

    actual, _ = expr.__teal__(avm2Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected


def test_ge_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.Ge(pt.Int(1), pt.Txn.receiver())

    with pytest.raises(pt.TealTypeError):
        pt.Ge(pt.Txn.receiver(), pt.Int(1))


def test_get_bit_int():
    args = [pt.Int(3), pt.Int(1)]
    expr = pt.GetBit(args[0], args[1])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.int, 3),
            pt.TealOp(args[1], pt.Op.int, 1),
            pt.TealOp(expr, pt.Op.getbit),
        ]
    )

    actual, _ = expr.__teal__(avm3Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm2Options)


def test_get_bit_bytes():
    args = [pt.Bytes("base16", "0xFF"), pt.Int(1)]
    expr = pt.GetBit(args[0], args[1])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, "0xFF"),
            pt.TealOp(args[1], pt.Op.int, 1),
            pt.TealOp(expr, pt.Op.getbit),
        ]
    )

    actual, _ = expr.__teal__(avm3Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm2Options)


def test_get_bit_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.GetBit(pt.Int(3), pt.Bytes("index"))

    with pytest.raises(pt.TealTypeError):
        pt.GetBit(pt.Bytes("base16", "0xFF"), pt.Bytes("index"))


def test_get_byte():
    args = [pt.Bytes("base16", "0xFF"), pt.Int(0)]
    expr = pt.GetByte(args[0], args[1])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, "0xFF"),
            pt.TealOp(args[1], pt.Op.int, 0),
            pt.TealOp(expr, pt.Op.getbyte),
        ]
    )

    actual, _ = expr.__teal__(avm3Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm2Options)


def test_get_byte_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.GetByte(pt.Int(3), pt.Int(0))

    with pytest.raises(pt.TealTypeError):
        pt.GetBit(pt.Bytes("base16", "0xFF"), pt.Bytes("index"))


def test_b_add():
    args = [
        pt.Bytes("base16", "0xFFFFFFFFFFFFFFFFFF"),
        pt.Bytes("base16", "0xFFFFFFFFFFFFFFFFFE"),
    ]
    expr = pt.BytesAdd(args[0], args[1])
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, "0xFFFFFFFFFFFFFFFFFF"),
            pt.TealOp(args[1], pt.Op.byte, "0xFFFFFFFFFFFFFFFFFE"),
            pt.TealOp(expr, pt.Op.b_add),
        ]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm3Options)


def test_b_add_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.BytesAdd(pt.Int(2), pt.Txn.receiver())

    with pytest.raises(pt.TealTypeError):
        pt.BytesAdd(pt.Bytes("base16", "0xFF"), pt.Int(2))


def test_b_minus():
    args = [
        pt.Bytes("base16", "0xFFFFFFFFFFFFFFFFFF"),
        pt.Bytes("base16", "0xFFFFFFFFFFFFFFFFFE"),
    ]
    expr = pt.BytesMinus(args[0], args[1])
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, "0xFFFFFFFFFFFFFFFFFF"),
            pt.TealOp(args[1], pt.Op.byte, "0xFFFFFFFFFFFFFFFFFE"),
            pt.TealOp(expr, pt.Op.b_minus),
        ]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm3Options)


def test_b_minus_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.BytesMinus(pt.Int(2), pt.Txn.receiver())

    with pytest.raises(pt.TealTypeError):
        pt.BytesMinus(pt.Bytes("base16", "0xFF"), pt.Int(2))


def test_b_div():
    args = [pt.Bytes("base16", "0xFFFFFFFFFFFFFFFF00"), pt.Bytes("base16", "0xFF")]
    expr = pt.BytesDiv(args[0], args[1])
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, "0xFFFFFFFFFFFFFFFF00"),
            pt.TealOp(args[1], pt.Op.byte, "0xFF"),
            pt.TealOp(expr, pt.Op.b_div),
        ]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm3Options)


def test_b_div_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.BytesDiv(pt.Int(2), pt.Txn.receiver())

    with pytest.raises(pt.TealTypeError):
        pt.BytesDiv(pt.Bytes("base16", "0xFF"), pt.Int(2))


def test_b_mul():
    args = [pt.Bytes("base16", "0xFFFFFFFFFFFFFFFF"), pt.Bytes("base16", "0xFF")]
    expr = pt.BytesMul(args[0], args[1])
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, "0xFFFFFFFFFFFFFFFF"),
            pt.TealOp(args[1], pt.Op.byte, "0xFF"),
            pt.TealOp(expr, pt.Op.b_mul),
        ]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm3Options)


def test_b_mul_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.BytesMul(pt.Int(2), pt.Txn.receiver())

    with pytest.raises(pt.TealTypeError):
        pt.BytesMul(pt.Bytes("base16", "0xFF"), pt.Int(2))


def test_b_mod():
    args = [pt.Bytes("base16", "0xFFFFFFFFFFFFFFFFFF"), pt.Bytes("base16", "0xFF")]
    expr = pt.BytesMod(args[0], args[1])
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, "0xFFFFFFFFFFFFFFFFFF"),
            pt.TealOp(args[1], pt.Op.byte, "0xFF"),
            pt.TealOp(expr, pt.Op.b_mod),
        ]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm3Options)


def test_b_mod_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.BytesMod(pt.Int(2), pt.Txn.receiver())

    with pytest.raises(pt.TealTypeError):
        pt.BytesMod(pt.Bytes("base16", "0xFF"), pt.Int(2))


def test_b_and():
    args = [pt.Bytes("base16", "0xFFFFFFFFFFFFFFFFF0"), pt.Bytes("base16", "0xFF")]
    expr = pt.BytesAnd(args[0], args[1])
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, "0xFFFFFFFFFFFFFFFFF0"),
            pt.TealOp(args[1], pt.Op.byte, "0xFF"),
            pt.TealOp(expr, pt.Op.b_and),
        ]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm3Options)


def test_b_and_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.BytesAnd(pt.Int(2), pt.Txn.receiver())

    with pytest.raises(pt.TealTypeError):
        pt.BytesAnd(pt.Bytes("base16", "0xFF"), pt.Int(2))


def test_b_or():
    args = [pt.Bytes("base16", "0xFFFFFFFFFFFFFFFFF0"), pt.Bytes("base16", "0xFF")]
    expr = pt.BytesOr(args[0], args[1])
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, "0xFFFFFFFFFFFFFFFFF0"),
            pt.TealOp(args[1], pt.Op.byte, "0xFF"),
            pt.TealOp(expr, pt.Op.b_or),
        ]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm3Options)


def test_b_or_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.BytesOr(pt.Int(2), pt.Txn.receiver())

    with pytest.raises(pt.TealTypeError):
        pt.BytesOr(pt.Bytes("base16", "0xFF"), pt.Int(2))


def test_b_xor():
    args = [pt.Bytes("base16", "0xFFFFFFFFFFFFFFFFF0"), pt.Bytes("base16", "0xFF")]
    expr = pt.BytesXor(args[0], args[1])
    assert expr.type_of() == pt.TealType.bytes

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, "0xFFFFFFFFFFFFFFFFF0"),
            pt.TealOp(args[1], pt.Op.byte, "0xFF"),
            pt.TealOp(expr, pt.Op.b_xor),
        ]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm3Options)


def test_b_xor_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.BytesXor(pt.Int(2), pt.Txn.receiver())

    with pytest.raises(pt.TealTypeError):
        pt.BytesXor(pt.Bytes("base16", "0xFF"), pt.Int(2))


def test_b_eq():
    args = [
        pt.Bytes("base16", "0xFFFFFFFFFFFFFFFFFF"),
        pt.Bytes("base16", "0xFFFFFFFFFFFFFFFFFF"),
    ]
    expr = pt.BytesEq(args[0], args[1])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, "0xFFFFFFFFFFFFFFFFFF"),
            pt.TealOp(args[1], pt.Op.byte, "0xFFFFFFFFFFFFFFFFFF"),
            pt.TealOp(expr, pt.Op.b_eq),
        ]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm3Options)


def test_b_eq_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.BytesEq(pt.Int(2), pt.Txn.receiver())

    with pytest.raises(pt.TealTypeError):
        pt.BytesEq(pt.Bytes("base16", "0xFF"), pt.Int(2))


def test_b_neq():
    args = [
        pt.Bytes("base16", "0xFFFFFFFFFFFFFFFFFF"),
        pt.Bytes("base16", "0xFFFFFFFFFFFFFFFFFF"),
    ]
    expr = pt.BytesNeq(args[0], args[1])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, "0xFFFFFFFFFFFFFFFFFF"),
            pt.TealOp(args[1], pt.Op.byte, "0xFFFFFFFFFFFFFFFFFF"),
            pt.TealOp(expr, pt.Op.b_neq),
        ]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm3Options)


def test_b_neq_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.BytesNeq(pt.Int(2), pt.Txn.receiver())

    with pytest.raises(pt.TealTypeError):
        pt.BytesNeq(pt.Bytes("base16", "0xFF"), pt.Int(2))


def test_b_lt():
    args = [
        pt.Bytes("base16", "0xFFFFFFFFFFFFFFFFF0"),
        pt.Bytes("base16", "0xFFFFFFFFFFFFFFFFFF"),
    ]
    expr = pt.BytesLt(args[0], args[1])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, "0xFFFFFFFFFFFFFFFFF0"),
            pt.TealOp(args[1], pt.Op.byte, "0xFFFFFFFFFFFFFFFFFF"),
            pt.TealOp(expr, pt.Op.b_lt),
        ]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm3Options)


def test_b_lt_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.BytesLt(pt.Int(2), pt.Txn.receiver())

    with pytest.raises(pt.TealTypeError):
        pt.BytesLt(pt.Bytes("base16", "0xFF"), pt.Int(2))


def test_b_le():
    args = [
        pt.Bytes("base16", "0xFFFFFFFFFFFFFFFFF0"),
        pt.Bytes("base16", "0xFFFFFFFFFFFFFFFFFF"),
    ]
    expr = pt.BytesLe(args[0], args[1])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, "0xFFFFFFFFFFFFFFFFF0"),
            pt.TealOp(args[1], pt.Op.byte, "0xFFFFFFFFFFFFFFFFFF"),
            pt.TealOp(expr, pt.Op.b_le),
        ]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm3Options)


def test_b_le_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.BytesLe(pt.Int(2), pt.Txn.receiver())

    with pytest.raises(pt.TealTypeError):
        pt.BytesLe(pt.Bytes("base16", "0xFF"), pt.Int(2))


def test_b_gt():
    args = [
        pt.Bytes("base16", "0xFFFFFFFFFFFFFFFFFF"),
        pt.Bytes("base16", "0xFFFFFFFFFFFFFFFFF0"),
    ]
    expr = pt.BytesGt(args[0], args[1])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, "0xFFFFFFFFFFFFFFFFFF"),
            pt.TealOp(args[1], pt.Op.byte, "0xFFFFFFFFFFFFFFFFF0"),
            pt.TealOp(expr, pt.Op.b_gt),
        ]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm3Options)


def test_b_gt_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.BytesGt(pt.Int(2), pt.Txn.receiver())

    with pytest.raises(pt.TealTypeError):
        pt.BytesGt(pt.Bytes("base16", "0xFF"), pt.Int(2))


def test_b_ge():
    args = [
        pt.Bytes("base16", "0xFFFFFFFFFFFFFFFFFF"),
        pt.Bytes("base16", "0xFFFFFFFFFFFFFFFFF0"),
    ]
    expr = pt.BytesGe(args[0], args[1])
    assert expr.type_of() == pt.TealType.uint64

    expected = pt.TealSimpleBlock(
        [
            pt.TealOp(args[0], pt.Op.byte, "0xFFFFFFFFFFFFFFFFFF"),
            pt.TealOp(args[1], pt.Op.byte, "0xFFFFFFFFFFFFFFFFF0"),
            pt.TealOp(expr, pt.Op.b_ge),
        ]
    )

    actual, _ = expr.__teal__(avm4Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    assert actual == expected

    with pytest.raises(pt.TealInputError):
        expr.__teal__(avm3Options)


def test_b_ge_invalid():
    with pytest.raises(pt.TealTypeError):
        pt.BytesGe(pt.Int(2), pt.Txn.receiver())

    with pytest.raises(pt.TealTypeError):
        pt.BytesGe(pt.Bytes("base16", "0xFF"), pt.Int(2))


def test_extract_uint():
    for expression, op in (
        (pt.ExtractUint16, pt.Op.extract_uint16),
        (pt.ExtractUint32, pt.Op.extract_uint32),
        (pt.ExtractUint64, pt.Op.extract_uint64),
    ):
        args = [
            pt.Bytes("base16", "0xFFFFFFFFFFFFFFFFFF"),
            pt.Int(2),
        ]
        expr = expression(args[0], args[1])
        assert expr.type_of() == pt.TealType.uint64

        expected = pt.TealSimpleBlock(
            [
                pt.TealOp(args[0], pt.Op.byte, "0xFFFFFFFFFFFFFFFFFF"),
                pt.TealOp(args[1], pt.Op.int, 2),
                pt.TealOp(expr, op),
            ]
        )

        actual, _ = expr.__teal__(avm5Options)
        actual.addIncoming()
        actual = pt.TealBlock.NormalizeBlocks(actual)

        assert actual == expected

        with pytest.raises(pt.TealInputError):
            expr.__teal__(avm4Options)


def test_extract_uint_invalid():
    for expression in (pt.ExtractUint16, pt.ExtractUint32, pt.ExtractUint64):
        with pytest.raises(pt.TealTypeError):
            expression(pt.Int(2), pt.Txn.receiver())

        with pytest.raises(pt.TealTypeError):
            expression(pt.Bytes("base16", "0xFF"), pt.Txn.receiver())



from typing import TYPE_CHECKING, cast, Union

from pyteal.config import MAX_GROUP_SIZE

from pyteal.errors import TealInputError, verifyFieldVersion, verifyProgramVersion
from pyteal.ir import TealOp, Op, TealBlock
from pyteal.ast.expr import Expr
from pyteal.ast.txn import TxnExpr, TxnField, TxnObject, TxnaExpr

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class GitxnExpr(TxnExpr):
    """An expression that accesses an inner transaction field from an inner transaction in the last inner group."""

    def __init__(self, txnIndex: int, field: TxnField) -> None:
        super().__init__(Op.gitxn, "Gitxn", field)

        # Currently we do not have gitxns. Only gitxn with immediate transaction index supported.
        if type(txnIndex) is not int:
            raise TealInputError(
                "Invalid gitxn syntax with immediate transaction field {} and transaction index {}".format(
                    field, txnIndex
                )
            )

        self.txnIndex = txnIndex

    def __str__(self):
        return "({} {} {})".format(self.name, self.txnIndex, self.field.arg_name)

    def __teal__(self, options: "CompileOptions"):
        verifyFieldVersion(self.field.arg_name, self.field.min_version, options.version)

        verifyProgramVersion(
            Op.gitxn.min_version,
            options.version,
            "Program version too low to use gitxn",
        )
        op = TealOp(self, Op.gitxn, self.txnIndex, self.field.arg_name)
        return TealBlock.FromOp(options, op)


GitxnExpr.__module__ = "pyteal"


class GitxnaExpr(TxnaExpr):
    """An expression that accesses an inner transaction array field from an inner transaction in the last inner group."""

    def __init__(self, txnIndex: int, field: TxnField, index: Union[int, Expr]) -> None:
        super().__init__(Op.gitxna, Op.gitxnas, "Gitxna", field, index)

        if type(txnIndex) is not int:
            raise TealInputError(
                f"Invalid txnIndex type:  Expected int, but received {txnIndex}."
            )

        self.txnIndex = txnIndex

    def __str__(self):
        return "({} {} {} {})".format(
            self.name, self.txnIndex, self.field.arg_name, self.index
        )

    def __teal__(self, options: "CompileOptions"):
        verifyFieldVersion(self.field.arg_name, self.field.min_version, options.version)

        if type(self.index) is int:
            opToUse = Op.gitxna
        else:
            opToUse = Op.gitxnas

        verifyProgramVersion(
            opToUse.min_version,
            options.version,
            "Program version too low to use op {}".format(opToUse),
        )

        if type(self.index) is int:
            op = TealOp(self, opToUse, self.txnIndex, self.field.arg_name, self.index)
            return TealBlock.FromOp(options, op)
        op = TealOp(self, opToUse, self.txnIndex, self.field.arg_name)
        return TealBlock.FromOp(options, op, cast(Expr, self.index))


GitxnaExpr.__module__ = "pyteal"


class InnerTxnGroup:
    """Represents a group of inner transactions."""

    def __getitem__(self, txnIndex: int) -> TxnObject:
        if type(txnIndex) is not int:
            raise TealInputError(
                "Invalid gitxn syntax, immediate txn index must be int."
            )

        if txnIndex < 0 or txnIndex >= MAX_GROUP_SIZE:
            raise TealInputError(
                "Invalid Gtxn index {}, should be in [0, {})".format(
                    txnIndex, MAX_GROUP_SIZE
                )
            )

        return TxnObject(
            lambda field: GitxnExpr(txnIndex, field),
            lambda field, index: GitxnaExpr(txnIndex, field, index),
        )


InnerTxnGroup.__module__ = "pyteal"

Gitxn: InnerTxnGroup = InnerTxnGroup()

Gitxn.__module__ = "pyteal"



# This example is provided for informational purposes only and has not been audited for security.

from pyteal import *

"""Split Payment"""

tmpl_fee = Int(1000)
tmpl_rcv1 = Addr("6ZHGHH5Z5CTPCF5WCESXMGRSVK7QJETR63M3NY5FJCUYDHO57VTCMJOBGY")
tmpl_rcv2 = Addr("7Z5PWO2C6LFNQFGHWKSK5H47IQP5OJW2M3HA2QPXTY3WTNP5NU2MHBW27M")
tmpl_own = Addr("5MK5NGBRT5RL6IGUSYDIX5P7TNNZKRVXKT6FGVI6UVK6IZAWTYQGE4RZIQ")
tmpl_ratn = Int(1)
tmpl_ratd = Int(3)
tmpl_min_pay = Int(1000)
tmpl_timeout = Int(3000)


def split(
    tmpl_fee=tmpl_fee,
    tmpl_rcv1=tmpl_rcv1,
    tmpl_rcv2=tmpl_rcv2,
    tmpl_own=tmpl_own,
    tmpl_ratn=tmpl_ratn,
    tmpl_ratd=tmpl_ratd,
    tmpl_min_pay=tmpl_min_pay,
    tmpl_timeout=tmpl_timeout,
):

    split_core = And(
        Txn.type_enum() == TxnType.Payment,
        Txn.fee() < tmpl_fee,
        Txn.rekey_to() == Global.zero_address(),
    )

    split_transfer = And(
        Gtxn[0].sender() == Gtxn[1].sender(),
        Txn.close_remainder_to() == Global.zero_address(),
        Gtxn[0].receiver() == tmpl_rcv1,
        Gtxn[1].receiver() == tmpl_rcv2,
        Gtxn[0].amount()
        == ((Gtxn[0].amount() + Gtxn[1].amount()) * tmpl_ratn) / tmpl_ratd,
        Gtxn[0].amount() == tmpl_min_pay,
    )

    split_close = And(
        Txn.close_remainder_to() == tmpl_own,
        Txn.receiver() == Global.zero_address(),
        Txn.amount() == Int(0),
        Txn.first_valid() > tmpl_timeout,
    )

    split_program = And(
        split_core, If(Global.group_size() == Int(2), split_transfer, split_close)
    )

    return split_program


if __name__ == "__main__":
    print(compileTeal(split(), mode=Mode.Signature, version=2))
